# Copyright 2021 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Tests for `deterministic.py`."""

from absl.testing import absltest
from absl.testing import parameterized

import chex
from distrax._src.distributions import deterministic
from distrax._src.utils import equivalence
import jax.experimental
import jax.numpy as jnp
import numpy as np


class DeterministicTest(equivalence.EquivalenceTest):

  def setUp(self):
    super().setUp()
    self._init_distr_cls(deterministic.Deterministic)

  def test_loc(self):
    dist_params = {'loc': [0.1, 0.5, 1.5]}
    dist = self.distrax_cls(**dist_params)
    self.assertion_fn(rtol=1e-3)(dist.loc, dist_params['loc'])

  @parameterized.named_parameters(
      ('None', None),
      ('0.0', 0.0),
      ('0.1', 0.1))
  def test_atol(self, atol):
    dist_params = {'loc': np.asarray([0.1, 0.5, 1.5]), 'atol': atol}
    dist = self.distrax_cls(**dist_params)
    broadcasted_atol = np.zeros((3,)) if atol is None else atol * np.ones((3,))
    self.assertion_fn(rtol=1e-3)(dist.atol, broadcasted_atol)

  @parameterized.named_parameters(
      ('None', None),
      ('0.0', 0.0),
      ('0.1', 0.1))
  def test_rtol(self, rtol):
    dist_params = {'loc': np.asarray([0.1, 0.5, 1.5]), 'rtol': rtol}
    dist = self.distrax_cls(**dist_params)
    broadcasted_rtol = np.zeros((3,)) if rtol is None else rtol * np.ones((3,))
    self.assertion_fn(rtol=1e-3)(dist.rtol, broadcasted_rtol)

  @parameterized.named_parameters(
      ('atol_None_rtol_None', None, None),
      ('atol_0.1_rtol_None', 0.1, None),
      ('atol_None_rtol_0.1', None, 0.1),
      ('atol_0.05_rtol_0.1', 0.05, 0.1))
  def test_slack(self, atol, rtol):
    loc = np.asarray([0.1, 0.5, 1.5])
    target_value = (0 if atol is None else atol) + (
        0 if rtol is None else rtol) * np.abs(loc)
    dist_params = {'loc': loc, 'rtol': rtol, 'atol': atol}
    dist = self.distrax_cls(**dist_params)
    self.assertion_fn(rtol=1e-3)(dist.slack, target_value)

  def test_invalid_parameters(self):
    self._test_raises_error(
        dist_kwargs={'loc': 2., 'atol': np.array([0.1, 0.2])})
    self._test_raises_error(
        dist_kwargs={'loc': 2., 'rtol': np.array([0.1, 0.2])})

  @parameterized.named_parameters(
      ('1d', np.asarray([0., 1.])),
      ('2d', np.zeros((2, 3))),
  )
  def test_event_shape(self, loc):
    dist_params = {'loc': loc}
    super()._test_event_shape((), dist_params)

  @chex.all_variants
  @parameterized.named_parameters(
      ('1d, no shape', [0., 1.], ()),
      ('1d, int shape', [0., 1.], 1),
      ('1d, 1-tuple shape', [0., 1.], (1,)),
      ('1d, 2-tuple shape', [0., 1.], (2, 2)),
      ('2d, no shape', np.zeros((2, 3)), ()),
      ('2d, int shape', np.zeros((2, 3)), 1),
      ('2d, 1-tuple shape', np.zeros((2, 3)), (1,)),
      ('2d, 2-tuple shape', np.zeros((2, 3)), (5, 4)),
  )
  def test_sample_shape(self, loc, sample_shape):
    dist_params = {'loc': np.asarray(loc)}
    super()._test_sample_shape(
        dist_args=(),
        dist_kwargs=dist_params,
        sample_shape=sample_shape)

  @chex.all_variants
  @parameterized.named_parameters(
      ('int32', jnp.int32),
      ('int64', jnp.int64),
      ('float32', jnp.float32),
      ('float64', jnp.float64))
  def test_sample_dtype(self, dtype):
    with jax.experimental.enable_x64(dtype.dtype.itemsize == 8):
      dist = self.distrax_cls(loc=jnp.zeros((), dtype=dtype))
      samples = self.variant(dist.sample)(seed=self.key)
      self.assertEqual(samples.dtype, dist.dtype)
      chex.assert_type(samples, dtype)

  @chex.all_variants
  @parameterized.named_parameters(
      ('1d, no shape', [0., 1.], ()),
      ('1d, int shape', [0., 1.], 1),
      ('1d, 1-tuple shape', [0., 1.], (1,)),
      ('1d, 2-tuple shape', [0., 1.], (2, 2)),
      ('2d, no shape', np.zeros((2, 3)), ()),
      ('2d, int shape', np.zeros((2, 3)), 1),
      ('2d, 1-tuple shape', np.zeros((2, 3)), (1,)),
      ('2d, 2-tuple shape', np.zeros((2, 3)), (5, 4)),
  )
  def test_sample_and_log_prob(self, loc, sample_shape):
    dist_params = {'loc': np.asarray(loc)}
    super()._test_sample_and_log_prob(
        dist_args=(),
        dist_kwargs=dist_params,
        sample_shape=sample_shape,
        assertion_fn=self.assertion_fn(rtol=1e-3))

  @chex.all_variants
  @parameterized.named_parameters(
      ('log_prob', 'log_prob'),
      ('prob', 'prob'),
      ('cdf', 'cdf'),
      ('log_cdf', 'log_cdf'),
  )
  def test_method_with_inputs_at_loc(self, function_string):
    loc = np.asarray([0.1, -0.9, 5.1])
    dist_params = {'loc': loc}
    inputs = np.repeat(loc[None, :], 10, axis=0)
    super()._test_attribute(
        attribute_string=function_string,
        dist_kwargs=dist_params,
        call_args=(inputs,),
        assertion_fn=self.assertion_fn(rtol=1e-3))

  @chex.all_variants
  @parameterized.named_parameters(
      ('log_prob', 'log_prob'),
      ('prob', 'prob'),
      ('cdf', 'cdf'),
      ('log_cdf', 'log_cdf'),
  )
  def test_method_with_inputs_at_random_inputs(self, function_string):
    loc = np.asarray([0.1, -0.9, 5.1])
    dist_params = {'loc': loc}
    inputs = 0.1 * np.random.normal(size=(10,) + (len(loc),))
    super()._test_attribute(
        attribute_string=function_string,
        dist_kwargs=dist_params,
        call_args=(inputs,),
        assertion_fn=self.assertion_fn(rtol=1e-3))

  @chex.all_variants
  @parameterized.named_parameters(
      ('log_prob_stddev0', 'log_prob', 0.0, 0.05, 0.1),
      ('log_prob_stddev0.05', 'log_prob', 0.05, 0.05, 0.1),
      ('log_prob_stddev0.1', 'log_prob', 0.1, 0.05, 0.1),
      ('prob_stddev0', 'prob', 0.0, 0.05, 0.1),
      ('prob_stddev0.05', 'prob', 0.05, 0.05, 0.1),
      ('prob_stddev0.1', 'prob', 0.1, 0.05, 0.1),
      ('cdf_stddev0', 'cdf', 0.0, 0.05, 0.1),
      ('cdf_stddev0.05', 'cdf', 0.05, 0.05, 0.1),
      ('cdf_stddev0.1', 'cdf', 0.1, 0.05, 0.1),
      ('log_cdf_stddev0', 'log_cdf', 0.0, 0.05, 0.1),
      ('log_cdf_stddev0.05', 'log_cdf', 0.05, 0.05, 0.1),
      ('log_cdf_stddev0.1', 'log_cdf', 0.1, 0.05, 0.1),
  )
  def test_method_with_inputs_and_slack(self, function_string, inputs_stddev,
                                        atol, rtol):
    loc = np.asarray([[4., -1., 0.], [0.5, 0.1, -8.]])
    dist_params = {'loc': loc, 'atol': atol, 'rtol': rtol}
    inputs = loc[None, ...] + inputs_stddev * np.random.normal(
        size=(20,) + loc.shape)
    super()._test_attribute(
        attribute_string=function_string,
        dist_kwargs=dist_params,
        call_args=(inputs,),
        assertion_fn=self.assertion_fn(rtol=1e-3))

  @chex.all_variants(with_pmap=False)
  @parameterized.named_parameters(
      ('entropy', [0., 1.], 'entropy'),
      ('mean', [0., 1.], 'mean'),
      ('mode', [0., 1.], 'mode'),
      ('variance', [0., 1.], 'variance'),
      ('variance from rank-2 params', np.ones((2, 3)), 'variance'),
      ('stddev', [-1.], 'stddev'),
      ('stddev from rank-2 params', -np.ones((2, 3)), 'stddev'),
  )
  def test_method(self, distr_params, function_string):
    super()._test_attribute(
        attribute_string=function_string,
        dist_kwargs={'loc': np.asarray(distr_params)},
        assertion_fn=self.assertion_fn(rtol=1e-3))

  @chex.all_variants(with_pmap=False)
  @parameterized.named_parameters(
      ('kl distrax_to_distrax', 'kl_divergence', 'distrax_to_distrax'),
      ('kl distrax_to_tfp', 'kl_divergence', 'distrax_to_tfp'),
      ('kl tfp_to_distrax', 'kl_divergence', 'tfp_to_distrax'),
      ('cross-ent distrax_to_distrax', 'cross_entropy', 'distrax_to_distrax'),
      ('cross-ent distrax_to_tfp', 'cross_entropy', 'distrax_to_tfp'),
      ('cross-ent tfp_to_distrax', 'cross_entropy', 'tfp_to_distrax'))
  def test_with_two_distributions(self, function_string, mode_string):
    loc1 = np.random.randn(3)
    loc2 = np.stack([loc1, np.random.randn(3)], axis=0)
    super()._test_with_two_distributions(
        attribute_string=function_string,
        mode_string=mode_string,
        dist1_kwargs={
            'loc': loc1,
        },
        dist2_kwargs={
            'loc': loc2,
        },
        assertion_fn=self.assertion_fn(rtol=1e-3))

  def test_jittable(self):
    super()._test_jittable((np.array([0., 4., -1., 4.]),))

  @parameterized.named_parameters(
      ('single element', 2),
      ('range', slice(-1)),
      ('range_2', (slice(None), slice(-1))),
      ('ellipsis', (Ellipsis, -1)),
  )
  def test_slice(self, slice_):
    loc = jnp.array(np.random.randn(3, 4, 5))
    atol = jnp.array(np.random.randn(3, 4, 5))
    rtol = jnp.array(np.random.randn(3, 4, 5))
    dist = self.distrax_cls(loc=loc, atol=atol, rtol=rtol)
    self.assertion_fn(rtol=1e-3)(dist[slice_].loc, loc[slice_])
    self.assertion_fn(rtol=1e-3)(dist[slice_].atol, atol[slice_])
    self.assertion_fn(rtol=1e-3)(dist[slice_].rtol, rtol[slice_])

  def test_slice_different_parameterization(self):
    loc = jnp.array(np.random.randn(3, 4, 5))
    atol = jnp.array(np.random.randn(4, 5))
    rtol = jnp.array(np.random.randn(4, 5))
    dist = self.distrax_cls(loc=loc, atol=atol, rtol=rtol)
    self.assertion_fn(rtol=1e-3)(dist[0].loc, loc[0])
    self.assertion_fn(rtol=1e-3)(dist[0].atol, atol)  # Not slicing atol.
    self.assertion_fn(rtol=1e-3)(dist[0].rtol, rtol)  # Not slicing rtol.


if __name__ == '__main__':
  absltest.main()



# Copyright 2021 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Tests for `tfp_compatible_distribution.py`."""

from absl.testing import absltest
from absl.testing import parameterized

import chex
from distrax._src.distributions.categorical import Categorical
from distrax._src.distributions.independent import Independent
from distrax._src.distributions.laplace import Laplace
from distrax._src.distributions.mvn_diag import MultivariateNormalDiag
from distrax._src.distributions.normal import Normal
from distrax._src.distributions.tfp_compatible_distribution import tfp_compatible_distribution
from distrax._src.distributions.transformed import Transformed
from distrax._src.distributions.uniform import Uniform
import jax
import jax.numpy as jnp
import numpy as np
from tensorflow_probability.substrates import jax as tfp

tfb = tfp.bijectors
tfd = tfp.distributions


class TFPCompatibleDistributionNormal(parameterized.TestCase):
  """Tests for Normal distribution."""

  def setUp(self):
    super().setUp()
    self._sample_shape = (np.int32(10),)
    self._seed = 42
    self._key = jax.random.PRNGKey(self._seed)
    self.base_dist = Normal(loc=jnp.array([0., 0.]), scale=jnp.array([1., 1.]))
    self.values = jnp.array([1., -1.])
    self.distrax_second_dist = Normal(loc=-1., scale=0.8)
    self.tfp_second_dist = tfd.Normal(loc=-1., scale=0.8)

  def assertion_fn(self, rtol):
    return lambda x, y: np.testing.assert_allclose(x, y, rtol=rtol)

  @property
  def wrapped_dist(self):
    return tfp_compatible_distribution(self.base_dist)

  def test_event_shape(self):
    self.assertEqual(self.wrapped_dist.event_shape, self.base_dist.event_shape)

  def test_event_shape_types(self):
    wrapped_dist = tfp_compatible_distribution(self.distrax_second_dist)
    self.assertEqual(
        type(wrapped_dist.event_shape), type(self.tfp_second_dist.event_shape))
    self.assertEqual(
        type(wrapped_dist.event_shape_tensor()),
        type(self.tfp_second_dist.event_shape_tensor()))

  def test_batch_shape(self):
    self.assertEqual(self.wrapped_dist.batch_shape, self.base_dist.batch_shape)

  @chex.all_variants
  def test_sample(self):
    def sample_fn(key):
      return self.wrapped_dist.sample(seed=key, sample_shape=self._sample_shape)
    sample_fn = self.variant(sample_fn)
    self.assertion_fn(rtol=1e-4)(
        sample_fn(self._key),
        self.base_dist.sample(sample_shape=self._sample_shape, seed=self._key))

  def test_experimental_local_measure(self):
    samples = self.wrapped_dist.sample(seed=self._key)
    expected_log_prob = self.wrapped_dist.log_prob(samples)

    log_prob, space = self.wrapped_dist.experimental_local_measure(
        samples, backward_compat=True)
    self.assertion_fn(rtol=1e-4)(log_prob, expected_log_prob)
    self.assertIsInstance(space, tfp.experimental.tangent_spaces.FullSpace)

  @chex.all_variants(with_pmap=False)
  @parameterized.named_parameters(
      ('mean', 'mean'),
      ('mode', 'mode'),
      ('median', 'median'),
      ('stddev', 'stddev'),
      ('variance', 'variance'),
      ('entropy', 'entropy'),
  )
  def test_method(self, method):
    try:
      expected_result = self.variant(getattr(self.base_dist, method))()
    except NotImplementedError:
      return
    except AttributeError:
      return
    result = self.variant(getattr(self.wrapped_dist, method))()
    self.assertion_fn(rtol=1e-4)(result, expected_result)

  @chex.all_variants
  @parameterized.named_parameters(
      ('log_prob', 'log_prob'),
      ('prob', 'prob'),
      ('log_cdf', 'log_cdf'),
      ('cdf', 'cdf'),
  )
  def test_method_with_value(self, method):
    try:
      expected_result = self.variant(
          getattr(self.base_dist, method))(self.values)
    except NotImplementedError:
      return
    except AttributeError:
      return
    result = self.variant(getattr(self.wrapped_dist, method))(self.values)
    self.assertion_fn(rtol=1e-4)(result, expected_result)

  @chex.all_variants
  @parameterized.named_parameters(
      ('kl_divergence', 'kl_divergence'),
      ('cross_entropy', 'cross_entropy'),
  )
  def test_with_two_distributions(self, method):
    """Test methods of the form listed below.

      D(distrax_distrib || wrapped_distrib),
      D(wrapped_distrib || distrax_distrib),
      D(tfp_distrib || wrapped_distrib),
      D(wrapped_distrib || tfp_distrib).

    Args:
      method: the method name to be tested
    """
    try:
      expected_result1 = self.variant(
          getattr(self.distrax_second_dist, method))(self.base_distribution)
      expected_result2 = self.variant(
          getattr(self.base_distribution, method))(self.distrax_second_dist)
    except NotImplementedError:
      return
    except AttributeError:
      return
    distrax_result1 = self.variant(getattr(self.distrax_second_dist, method))(
        self.wrapped_dist)
    distrax_result2 = self.variant(getattr(self.wrapped_dist, method))(
        self.distrax_second_dist)
    tfp_result1 = self.variant(getattr(self.tfp_second_dist, method))(
        self.wrapped_dist)
    tfp_result2 = self.variant(getattr(self.wrapped_dist, method))(
        self.tfp_second_dist)
    self.assertion_fn(rtol=1e-4)(distrax_result1, expected_result1)
    self.assertion_fn(rtol=1e-4)(distrax_result2, expected_result2)
    self.assertion_fn(rtol=1e-4)(tfp_result1, expected_result1)
    self.assertion_fn(rtol=1e-4)(tfp_result2, expected_result2)


class TFPCompatibleDistributionMvnNormal(TFPCompatibleDistributionNormal):
  """Tests for multivariate normal distribution."""

  def setUp(self):
    super().setUp()
    self.base_dist = MultivariateNormalDiag(loc=jnp.array([0., 1.]))
    self.values = jnp.array([1., -1.])
    self.distrax_second_dist = MultivariateNormalDiag(
        loc=jnp.array([-1., 0.]), scale_diag=jnp.array([0.8, 1.2]))
    self.tfp_second_dist = tfd.MultivariateNormalDiag(
        loc=jnp.array([-1., 0.]), scale_diag=jnp.array([0.8, 1.2]))


class TFPCompatibleDistributionCategorical(TFPCompatibleDistributionNormal):
  """Tests for categorical distribution."""

  def setUp(self):
    super().setUp()
    self.base_dist = Categorical(logits=jnp.array([0., -1., 1.]))
    self.values = jnp.array([0, 1, 2])
    self.distrax_second_dist = Categorical(probs=jnp.array([0.2, 0.2, 0.6]))
    self.tfp_second_dist = tfd.Categorical(probs=jnp.array([0.2, 0.2, 0.6]))


class TFPCompatibleDistributionTransformed(TFPCompatibleDistributionNormal):
  """Tests for transformed distributions."""

  def setUp(self):
    super().setUp()
    self.base_dist = Transformed(
        distribution=Normal(loc=0., scale=1.),
        bijector=tfb.Exp())
    self.values = jnp.array([0., 1., 2.])
    self.distrax_second_dist = Transformed(
        distribution=Normal(loc=0.5, scale=0.8),
        bijector=tfb.Exp())
    self.tfp_second_dist = tfd.TransformedDistribution(
        distribution=tfd.Normal(loc=0.5, scale=0.8),
        bijector=tfb.Exp())


class TfpMetaDistributionsWithWrappedBaseDistribution(parameterized.TestCase):
  """Tests for meta distributions (with wrappper base distr)."""

  def setUp(self):
    super().setUp()
    self._sample_shape = (np.int32(10),)
    self._seed = 42
    self._key = jax.random.PRNGKey(self._seed)

  def assertion_fn(self, rtol):
    return lambda x, y: np.testing.assert_allclose(x, y, rtol=rtol)

  def test_with_independent(self):
    base_dist = Normal(loc=jnp.array([0., 0.]), scale=jnp.array([1., 1.]))
    wrapped_dist = tfp_compatible_distribution(base_dist)

    meta_dist = tfd.Independent(wrapped_dist, 1, validate_args=True)
    samples = meta_dist.sample((), self._key)
    log_prob = meta_dist.log_prob(samples)

    distrax_meta_dist = Independent(base_dist, 1)
    expected_log_prob = distrax_meta_dist.log_prob(samples)

    self.assertion_fn(rtol=1e-4)(log_prob, expected_log_prob)

  def test_with_transformed_distribution(self):
    base_dist = Normal(loc=jnp.array([0., 0.]), scale=jnp.array([1., 1.]))
    wrapped_dist = tfp_compatible_distribution(base_dist)

    meta_dist = tfd.TransformedDistribution(
        distribution=wrapped_dist, bijector=tfb.Exp(), validate_args=True)
    samples = meta_dist.sample(seed=self._key)
    log_prob = meta_dist.log_prob(samples)

    distrax_meta_dist = Transformed(
        distribution=base_dist, bijector=tfb.Exp())
    expected_log_prob = distrax_meta_dist.log_prob(samples)

    self.assertion_fn(rtol=1e-4)(log_prob, expected_log_prob)

  def test_with_sample(self):
    base_dist = Normal(0., 1.)
    wrapped_dist = tfp_compatible_distribution(base_dist)
    meta_dist = tfd.Sample(
        wrapped_dist, sample_shape=[1, 3], validate_args=True)
    meta_dist.log_prob(meta_dist.sample(2, seed=self._key))

  def test_with_joint_distribution_named_auto_batched(self):
    def laplace(a, b):
      return tfp_compatible_distribution(Laplace(a * jnp.ones((2, 1)), b))
    meta_dist = tfd.JointDistributionNamedAutoBatched({
        'a': tfp_compatible_distribution(Uniform(2. * jnp.ones(3), 4.)),
        'b': tfp_compatible_distribution(Uniform(2. * jnp.ones(3), 4.)),
        'c': laplace}, validate_args=True)
    meta_dist.log_prob(meta_dist.sample(4, seed=self._key))

  def test_with_joint_distribution_coroutine_auto_batched(self):

    def model_fn():
      a = yield tfp_compatible_distribution(Uniform(2. * jnp.ones(3), 4.),
                                            name='a')
      b = yield tfp_compatible_distribution(Uniform(2. * jnp.ones(3), 4.),
                                            name='b')
      yield tfp_compatible_distribution(Laplace(a * jnp.ones((2, 1)), b),
                                        name='c')

    meta_dist = tfd.JointDistributionCoroutineAutoBatched(
        model_fn, validate_args=True)
    meta_dist.log_prob(meta_dist.sample(7, seed=self._key))


class TFPCompatibleDistributionSlicing(parameterized.TestCase):
  """Class to test the `getitem` method."""

  def assertion_fn(self, rtol):
    return lambda x, y: np.testing.assert_allclose(x, y, rtol=rtol)

  @parameterized.named_parameters(
      ('single element', 2),
      ('range', slice(-1)),
      ('range_2', (slice(None), slice(-1))),
      ('ellipsis', (Ellipsis, -1)),
  )
  def test_slice(self, slice_):
    loc = np.random.randn(3, 4, 5)
    base_dist = Normal(loc=loc, scale=1.)
    dist = tfp_compatible_distribution(base_dist)
    sliced_dist = dist[slice_]
    self.assertIsInstance(sliced_dist, base_dist.__class__)
    self.assertIsInstance(sliced_dist.batch_shape, tfp.tf2jax.TensorShape)
    self.assertTrue(sliced_dist.allow_nan_stats)
    self.assertion_fn(rtol=1e-4)(sliced_dist.loc, loc[slice_])


if __name__ == '__main__':
  absltest.main()



# Copyright 2021 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Tests for `scalar_affine.py`."""

from absl.testing import absltest
from absl.testing import parameterized

import chex
from distrax._src.bijectors import scalar_affine
import jax
import jax.numpy as jnp
import numpy as np


class ScalarAffineTest(parameterized.TestCase):

  def test_properties(self):
    bij = scalar_affine.ScalarAffine(shift=0., scale=1.)
    self.assertTrue(bij.is_constant_jacobian)
    self.assertTrue(bij.is_constant_log_det)
    np.testing.assert_allclose(bij.shift, 0.)
    np.testing.assert_allclose(bij.scale, 1.)
    np.testing.assert_allclose(bij.log_scale, 0.)

  def test_raises_if_both_scale_and_log_scale_are_specified(self):
    with self.assertRaises(ValueError):
      scalar_affine.ScalarAffine(shift=0., scale=1., log_scale=0.)

  @chex.all_variants
  def test_shapes_are_correct(self):
    k1, k2, k3, k4 = jax.random.split(jax.random.PRNGKey(42), 4)
    x = jax.random.normal(k1, (2, 3, 4, 5))
    shift = jax.random.normal(k2, (4, 5))
    scale = jax.random.uniform(k3, (3, 4, 5)) + 0.1
    log_scale = jax.random.normal(k4, (3, 4, 5))
    bij_no_scale = scalar_affine.ScalarAffine(shift)
    bij_with_scale = scalar_affine.ScalarAffine(shift, scale=scale)
    bij_with_log_scale = scalar_affine.ScalarAffine(shift, log_scale=log_scale)
    for bij in [bij_no_scale, bij_with_scale, bij_with_log_scale]:
      # Forward methods.
      y, logdet = self.variant(bij.forward_and_log_det)(x)
      self.assertEqual(y.shape, (2, 3, 4, 5))
      self.assertEqual(logdet.shape, (2, 3, 4, 5))
      # Inverse methods.
      x, logdet = self.variant(bij.inverse_and_log_det)(y)
      self.assertEqual(x.shape, (2, 3, 4, 5))
      self.assertEqual(logdet.shape, (2, 3, 4, 5))

  @chex.all_variants
  def test_forward_methods_are_correct(self):
    key = jax.random.PRNGKey(42)
    x = jax.random.normal(key, (2, 3, 4, 5))
    bij_no_scale = scalar_affine.ScalarAffine(shift=3.)
    bij_with_scale = scalar_affine.ScalarAffine(shift=3., scale=1.)
    bij_with_log_scale = scalar_affine.ScalarAffine(shift=3., log_scale=0.)
    for bij in [bij_no_scale, bij_with_scale, bij_with_log_scale]:
      y, logdet = self.variant(bij.forward_and_log_det)(x)
      np.testing.assert_allclose(y, x + 3., atol=1e-8)
      np.testing.assert_allclose(logdet, 0., atol=1e-8)

  @chex.all_variants
  def test_inverse_methods_are_correct(self):
    k1, k2, k3, k4 = jax.random.split(jax.random.PRNGKey(42), 4)
    x = jax.random.normal(k1, (2, 3, 4, 5))
    shift = jax.random.normal(k2, (4, 5))
    scale = jax.random.uniform(k3, (3, 4, 5)) + 0.1
    log_scale = jax.random.normal(k4, (3, 4, 5))
    bij_no_scale = scalar_affine.ScalarAffine(shift)
    bij_with_scale = scalar_affine.ScalarAffine(shift, scale=scale)
    bij_with_log_scale = scalar_affine.ScalarAffine(shift, log_scale=log_scale)
    for bij in [bij_no_scale, bij_with_scale, bij_with_log_scale]:
      y, logdet_fwd = self.variant(bij.forward_and_log_det)(x)
      x_rec, logdet_inv = self.variant(bij.inverse_and_log_det)(y)
      np.testing.assert_allclose(x_rec, x, atol=1e-5)
      np.testing.assert_allclose(logdet_fwd, -logdet_inv, atol=3e-6)

  @chex.all_variants
  def test_composite_methods_are_consistent(self):
    k1, k2, k3, k4 = jax.random.split(jax.random.PRNGKey(42), 4)
    bij = scalar_affine.ScalarAffine(
        shift=jax.random.normal(k1, (4, 5)),
        log_scale=jax.random.normal(k2, (4, 5)))
    # Forward methods.
    x = jax.random.normal(k3, (2, 3, 4, 5))
    y1 = self.variant(bij.forward)(x)
    logdet1 = self.variant(bij.forward_log_det_jacobian)(x)
    y2, logdet2 = self.variant(bij.forward_and_log_det)(x)
    np.testing.assert_allclose(y1, y2, atol=1e-12)
    np.testing.assert_allclose(logdet1, logdet2, atol=1e-12)
    # Inverse methods.
    y = jax.random.normal(k4, (2, 3, 4, 5))
    x1 = self.variant(bij.inverse)(y)
    logdet1 = self.variant(bij.inverse_log_det_jacobian)(y)
    x2, logdet2 = self.variant(bij.inverse_and_log_det)(y)
    np.testing.assert_allclose(x1, x2, atol=1e-12)
    np.testing.assert_allclose(logdet1, logdet2, atol=1e-12)

  @chex.all_variants
  @parameterized.parameters(
      ((5,), (5,), (5,)),
      ((5,), (5,), ()),
      ((5,), (), (5,)),
      ((), (5,), (5,)),
      ((), (), (5,)),
      ((), (5,), ()),
      ((5,), (), ()),
  )
  def test_batched_parameters(self, scale_batch_shape, shift_batch_shape,
                              input_batch_shape):
    k1, k2, k3 = jax.random.split(jax.random.PRNGKey(42), 3)
    log_scale = jax.random.normal(k1, scale_batch_shape)
    shift = jax.random.normal(k2, shift_batch_shape)
    bijector = scalar_affine.ScalarAffine(shift, log_scale=log_scale)

    x = jax.random.normal(k3, input_batch_shape)
    y, logdet_fwd = self.variant(bijector.forward_and_log_det)(x)
    z, logdet_inv = self.variant(bijector.inverse_and_log_det)(x)

    output_batch_shape = jnp.broadcast_arrays(log_scale, shift, x)[0].shape

    self.assertEqual(y.shape, output_batch_shape)
    self.assertEqual(z.shape, output_batch_shape)
    self.assertEqual(logdet_fwd.shape, output_batch_shape)
    self.assertEqual(logdet_inv.shape, output_batch_shape)

    log_scale = jnp.broadcast_to(log_scale, output_batch_shape).flatten()
    shift = jnp.broadcast_to(shift, output_batch_shape).flatten()
    x = jnp.broadcast_to(x, output_batch_shape).flatten()
    y = y.flatten()
    z = z.flatten()
    logdet_fwd = logdet_fwd.flatten()
    logdet_inv = logdet_inv.flatten()

    for i in range(np.prod(output_batch_shape)):
      bijector = scalar_affine.ScalarAffine(shift[i], jnp.exp(log_scale[i]))
      this_y, this_logdet_fwd = self.variant(bijector.forward_and_log_det)(x[i])
      this_z, this_logdet_inv = self.variant(bijector.inverse_and_log_det)(x[i])
      np.testing.assert_allclose(this_y, y[i], atol=1e-7)
      np.testing.assert_allclose(this_z, z[i], atol=1e-5)
      np.testing.assert_allclose(this_logdet_fwd, logdet_fwd[i], atol=1e-4)
      np.testing.assert_allclose(this_logdet_inv, logdet_inv[i], atol=1e-4)

  def test_jittable(self):
    @jax.jit
    def f(x, b):
      return b.forward(x)

    bijector = scalar_affine.ScalarAffine(0, 1)
    x = np.zeros(())
    f(x, bijector)


if __name__ == '__main__':
  jax.config.update('jax_threefry_partitionable', False)
  absltest.main()



# Copyright 2022 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Von Mises distribution."""

import functools
import math
from typing import cast, Sequence, Tuple, Union

import chex
from distrax._src.distributions import distribution
from distrax._src.distributions import normal
from distrax._src.utils import conversion
import jax
import jax.numpy as jnp
import numpy as np
from tensorflow_probability.substrates import jax as tfp

tfd = tfp.distributions

Array = chex.Array
Numeric = chex.Numeric
PRNGKey = chex.PRNGKey
IntLike = Union[int, np.int16, np.int32, np.int64]
EventT = distribution.EventT


class VonMises(distribution.Distribution):
  """The von Mises distribution over angles.

  The von Mises distribution is a distribution over angles. It is the maximum
  entropy distribution on the space of angles, given a circular mean and a
  circular variance.

  In this implementation, the distribution is defined over the range [-pi, pi),
  with all samples in this interval and the CDF is constant outside this
  interval. Do note that the prob and log_prob also accept values outside of
  [-pi, pi) and will return values as if they are inside the interval.

  When `concentration=0`, this distribution becomes the uniform distribution
  over the interval [-pi, pi). When the concentration goes to infinity, this
  distribution approximates a Normal distribution.

  #### Details

  The probability density function (pdf) of this distribution is,

  ```none
  pdf(x; loc, concentration) = exp(concentration * cos(x - loc))
  / (2 * pi * I_0 (concentration))
  ```

  where:
  * `I_0` is the zeroth order modified Bessel function;
  * `loc` the circular mean of the distribution, a scalar in radians.
    It can take arbitrary values, also outside of [-pi, pi).
  * `concentration >= 0` is the concentration parameter. It is the
    analogue to 1/sigma of the Normal distribution.

  #### Examples

  Examples of initialization of this distribution.

  ```python
  # Create a batch of two von Mises distributions.
  dist = distrax.VonMises(loc=[1.0, 2.0], concentration=[3.0, 4.0])
  dist.sample(sample_shape=(3,), seed=0)  # Sample of shape [3, 2]
  ```

  Arguments are broadcast when possible.

  ```python
  dist = distrax.VonMises(loc=1.0, concentration=[3.0, 4.0])

  # Evaluating the pdf of both distributions on the point 3.0 returns a length 2
  # tensor.
  dist.prob(3.0)
  ```
  """
  equiv_tfp_cls = tfd.VonMises

  def __init__(self, loc: Numeric, concentration: Numeric):
    super().__init__()
    self._loc = conversion.as_float_array(loc)
    self._concentration = conversion.as_float_array(concentration)
    self._batch_shape = jax.lax.broadcast_shapes(
        self._loc.shape, self._concentration.shape
    )

  @property
  def loc(self) -> Array:
    """The circular mean of the distribution."""
    return jnp.broadcast_to(self._loc, self.batch_shape)

  @property
  def concentration(self) -> Array:
    """The concentration of the distribution."""
    return jnp.broadcast_to(self._concentration, self.batch_shape)

  @property
  def event_shape(self) -> Tuple[int, ...]:
    """Shape of event of distribution samples."""
    return ()

  @property
  def batch_shape(self) -> Tuple[int, ...]:
    """Shape of batch of distribution samples."""
    return self._batch_shape

  def mean(self) -> Array:
    """The circular mean of the distribution."""
    return self.loc

  def variance(self) -> Array:
    """The circular variance of the distribution."""
    conc = self._concentration
    return 1. - jax.scipy.special.i1e(conc) / jax.scipy.special.i0e(conc)

  def prob(self, value: EventT) -> Array:
    """The probability of value under the distribution."""
    conc = self._concentration
    unnormalized_prob = jnp.exp(conc * (jnp.cos(value - self._loc) - 1.))
    normalization = (2. * math.pi) * jax.scipy.special.i0e(conc)
    return unnormalized_prob / normalization

  def log_prob(self, value: EventT) -> Array:
    """The logarithm of the probability of value under the distribution."""
    conc = self._concentration
    i_0 = jax.scipy.special.i0(conc)
    return (
        conc * jnp.cos(value - self._loc) - math.log(2 * math.pi) - jnp.log(i_0)
    )

  def _sample_n(self, key: PRNGKey, n: int) -> Array:
    """Returns `n` samples in [-pi, pi)."""
    out_shape = (n,) + self.batch_shape
    conc = self._concentration
    dtype = jnp.result_type(self._loc, self._concentration)
    sample = _von_mises_sample(out_shape, conc, key, dtype) + self._loc
    return _convert_angle_to_standard(sample)

  def entropy(self) -> Array:
    """Returns the entropy."""
    conc = self._concentration
    i0e = jax.scipy.special.i0e(conc)
    i1e = jax.scipy.special.i1e(conc)
    return conc * (1 - i1e / i0e) + math.log(2 * math.pi) + jnp.log(i0e)

  def mode(self) -> Array:
    """The mode of the distribution."""
    return self.mean()

  def cdf(self, value: EventT) -> Array:
    """The CDF of `value` under the distribution.

    Note that the CDF takes values of 0. or 1. for values outside of
    [-pi, pi). Note that this behaviour is different from
    `tensorflow_probability.VonMises` or `scipy.stats.vonmises`.
    Args:
      value: the angle evaluated under the distribution.
    Returns:
      the circular CDF of value.
    """
    dtype = jnp.result_type(value, self._loc, self._concentration)
    loc = _convert_angle_to_standard(self._loc)
    return jnp.clip(
        _von_mises_cdf(value - loc, self._concentration, dtype)
        - _von_mises_cdf(-math.pi - loc, self._concentration, dtype),
        min=0.0,
        max=1.0,
    )

  def log_cdf(self, value: EventT) -> Array:
    """See `Distribution.log_cdf`."""
    return jnp.log(self.cdf(value))

  def survival_function(self, value: EventT) -> Array:
    """See `Distribution.survival_function`."""
    dtype = jnp.result_type(value, self._loc, self._concentration)
    loc = _convert_angle_to_standard(self._loc)
    return jnp.clip(
        _von_mises_cdf(math.pi - loc, self._concentration, dtype)
        - _von_mises_cdf(value - loc, self._concentration, dtype),
        min=0.0,
        max=1.0,
    )

  def log_survival_function(self, value: EventT) -> Array:
    """See `Distribution.log_survival_function`."""
    return jnp.log(self.survival_function(value))

  def __getitem__(self, index) -> 'VonMises':
    index = distribution.to_batch_shape_index(self.batch_shape, index)
    return VonMises(
        loc=self.loc[index],
        concentration=self.concentration[index],
    )


def _convert_angle_to_standard(angle: Array) -> Array:
  """Converts angle in radians to representation between [-pi, pi)."""
  num_periods = jnp.round(angle / (2 * math.pi))
  rep = angle - (2 * math.pi) * num_periods
  return rep


@functools.partial(jax.custom_jvp, nondiff_argnums=(0, 2, 3))
def _von_mises_sample(
    shape: Union[IntLike, Sequence[IntLike]],
    concentration: Array,
    seed: PRNGKey,
    dtype: jnp.dtype,
) -> Array:
  """Rejection sample from the standard von Mises which has loc=0."""
  concentration = jnp.asarray(concentration, dtype=dtype)
  s_concentration_cutoff_dict = {
      jnp.float16.dtype: 1.8e-1,
      jnp.float32.dtype: 2e-2,
      jnp.float64.dtype: 1.2e-4,
  }
  s_concentration_cutoff = s_concentration_cutoff_dict[dtype]
  use_exact = concentration > s_concentration_cutoff
  # Avoid NaN's, even when not used later.
  conc = jnp.where(use_exact, concentration, 1.)
  r = 1. + jnp.sqrt(1 + 4 * jnp.square(conc))
  rho = (r - jnp.sqrt(2. * r)) / (2 * conc)
  s_exact = (1. + jnp.square(rho)) / (2. * rho)
  s_approximate = 1.0 / jnp.clip(concentration, min=1e-7)
  s = jnp.where(use_exact, s_exact, s_approximate)

  def loop_body(arg):
    done, u_in, w, seed, count = arg
    del u_in
    u_seed, v_seed, next_seed = jax.random.split(seed, 3)
    u = jax.random.uniform(
        u_seed, shape=shape, dtype=dtype, minval=-1., maxval=1.
    )
    z = jnp.cos(math.pi * u)
    w = jnp.where(done, w, (1 + s * z) / (s + z))
    y = concentration * (s - w)
    v = jax.random.uniform(v_seed, shape, dtype=dtype, minval=0., maxval=1.)
    # Use `logical_not` to accept all "nan" samples.
    accept = jnp.logical_not(y * jnp.exp(1 - y) < v)
    return jnp.logical_or(accept, done), u, w, next_seed, count + 1

  def loop_cond(arg):
    done, u_in, w, seed, count = arg
    del u_in, w, seed
    # The rejection sampling is actually very efficient. With the worst
    # concentration, about half of the samples are rejected. So only
    # 1 in 1.51e25 samples will ever hit this counter on the worst possible
    # concentration, which is a point way beyond numerical accuracy anyway.
    return jnp.logical_and(jnp.any(jnp.logical_not(done)), count < 100)

  _, u, w, _, _ = jax.lax.while_loop(
      loop_cond,
      loop_body,
      init_val=(
          jnp.zeros(shape, dtype=jnp.bool_),
          jnp.zeros(shape, dtype=dtype),
          jnp.zeros(shape, dtype=dtype),
          seed,
          0
      )
  )
  return jnp.sign(u) * jnp.arccos(jnp.clip(w, min=-1.0, max=1.0))


# Since rejection sampling does not permit autodiff, add an analytic gradient.
@_von_mises_sample.defjvp
def _von_mises_sample_jvp(
    shape: Union[IntLike, Sequence[IntLike]],
    seed: PRNGKey,
    dtype: jnp.dtype,
    primals: Tuple[Array],
    tangents: Tuple[Array],
) -> Tuple[Array, Array]:
  """Returns the jvp of the von Mises sample operation."""
  concentration, = primals
  dconcentration, = tangents

  concentration = jnp.clip(concentration, min=1e-7)
  samples = _von_mises_sample(shape, concentration, seed, dtype)
  vectorized_grad_cdf = jnp.vectorize(
      jax.grad(_von_mises_cdf, argnums=1),
      signature='(),()->()',
      excluded=(2,),
  )
  dcdf_dconcentration = vectorized_grad_cdf(samples, concentration, dtype)

  inv_prob = jnp.exp(-concentration * (jnp.cos(samples) - 1.)) * (
      (2. * math.pi) * jax.scipy.special.i0e(concentration)
  )
  dcdf_dconcentration = cast(chex.Array, dcdf_dconcentration)
  dsamples = dconcentration * (-inv_prob * dcdf_dconcentration)
  return samples, dsamples


@functools.partial(jax.custom_jvp, nondiff_argnums=(2,))
def _von_mises_cdf(
    value: Array,
    concentration: Array,
    dtype: jnp.dtype,
) -> Array:
  """Returns the cumulative density function (CDF) of von Mises distribution.

  Denote the density of VonMises(loc=0, concentration=concentration) by p(t).
  The CDF at the point x is defined as int_{-pi}^x p(t) dt when x is in the
  interval [-pi, pi]. Below -pi, the CDF is zero, above pi, it is one.

  The CDF is not available in closed form. Instead, we use the method [1]
  which uses either a series expansion or a Normal approximation, depending on
  the value of concentration.

  We also compute the derivative of the CDF w.r.t. both x and concentration
  using the method described in [2].

  Args:
    value: The point at which to evaluate the CDF.
    concentration: The concentration parameter of the von Mises distribution.
    dtype: Type of the return value.

  Returns:
    The value of the CDF computed elementwise.

  References:
    [1] G. Hill "Algorithm 518: Incomplete Bessel Function I_0. The Von Mises
    Distribution." ACM Transactions on Mathematical Software, 1977.
    [2] Figurnov, M., Mohamed, S. and Mnih, A., "Implicit reparameterization
    gradients." Advances in Neural Information Processing Systems, 31, 2018.
  """
  primals = (value, concentration)
  tangents = (jnp.zeros_like(value), jnp.zeros_like(concentration))
  primal_out, _ = _von_mises_cdf_jvp(dtype, primals, tangents)
  return primal_out


# Use a custom jvp to increase numerical accuracy.
@_von_mises_cdf.defjvp
def _von_mises_cdf_jvp(
    dtype: jnp.dtype,
    primals: Tuple[Array, Array],
    tangents: Tuple[Array, Array],
):
  """Returns the jvp CDF of a von Mises."""
  x, concentration = primals
  dx, dconcentration = tangents
  num_periods = jnp.round(x / (2 * math.pi))
  x = x - (2 * math.pi) * num_periods

  # This is the cutoff-concentration for choosing between the two numerical
  # recipes for computing the CDF. For concentrations larger than
  # `concentration_cutoff`, a Normal approximation is used.
  concentration_cutoff = 10.5

  cdf_series, dcdf_dconcentration_series = _von_mises_cdf_series(
      x, concentration, dtype
  )
  cdf_normal, dcdf_dconcentration_normal = _von_mises_cdf_normal(
      x, concentration, dtype
  )
  use_series = concentration < concentration_cutoff
  cdf = jnp.where(use_series, cdf_series, cdf_normal)
  cdf = cdf + num_periods
  dcdf_dconcentration = jnp.where(
      use_series,
      dcdf_dconcentration_series,
      dcdf_dconcentration_normal,
  )

  prob = jnp.exp(concentration * (jnp.cos(x) - 1.)) / (
      (2. * math.pi) * jax.scipy.special.i0e(concentration)
  )

  return cdf, dconcentration * dcdf_dconcentration + dx * prob


def _von_mises_cdf_series(
    value: Array,
    concentration: Array,
    dtype: jnp.dtype,
    num_terms: int = 20,
) -> Tuple[Array, Array]:
  """Computes the CDF based on a series of `num_terms` terms."""
  rn = jnp.zeros_like(value, dtype=dtype)
  drn_dconcentration = jnp.zeros_like(value, dtype=dtype)
  vn = jnp.zeros_like(value, dtype=dtype)
  dvn_dconcentration = jnp.zeros_like(value, dtype=dtype)

  for n in range(num_terms, 0, -1):
    denominator = 2. * n / concentration + rn
    ddenominator_dk = -2. * n / jnp.square(concentration) + drn_dconcentration
    rn = 1. / denominator
    drn_dconcentration = -ddenominator_dk / jnp.square(denominator)
    multiplier = jnp.sin(n * value) / n + vn
    vn = rn * multiplier
    dvn_dconcentration = (
        drn_dconcentration * multiplier + rn * dvn_dconcentration
    )
  cdf = .5 + value / (2. * math.pi) + vn / math.pi
  dcdf_dconcentration = dvn_dconcentration / math.pi
  # Clip the result to [0, 1].
  cdf_clipped = jnp.clip(cdf, 0., 1.)
  # The clipped values do not depend on concentration anymore, so set their
  # derivative to zero.
  dcdf_dconcentration = (
      dcdf_dconcentration * jnp.logical_and(cdf >= 0., cdf <= 1.)
  )
  return cdf_clipped, dcdf_dconcentration


def _von_mises_cdf_normal(
    value: Array,
    concentration: Array,
    dtype: jnp.dtype,
) -> Tuple[Array, Array]:
  """Computes the CDF, based on a Normal approximation to the von Mises."""
  def cdf_func(value, concentration):
    """A helper function that is passed to value_and_gradient."""
    # z is an "almost Normally distributed" random variable.
    z = (
        jnp.sqrt(2. / math.pi) /
        jax.scipy.special.i0e(concentration) * jnp.sin(.5 * value)
    )
    # This is a correction described in [1].
    # It reduces the error of the Normal approximation.
    z2 = jnp.square(z)
    z3 = z2 * z
    z4 = jnp.square(z2)
    c = 24. * concentration
    c1 = 56.

    xi = z - z3 / jnp.square(
        (c - 2. * z2 - 16.) / 3. -
        (z4 + (7. / 4.) * z2 + 167. / 2.) / (c - c1 - z2 + 3.)
    )
    distrib = normal.Normal(
        loc=jnp.array(0., dtype),
        scale=jnp.array(1., dtype)
    )
    return distrib.cdf(xi)

  vectorized_cdf_with_grad = jnp.vectorize(
      jax.value_and_grad(cdf_func, argnums=1),
      signature='(),()->(),()',
  )
  return vectorized_cdf_with_grad(value, concentration)


def _kl_divergence_vonmises_vonmises(
    dist1: Union[VonMises, tfd.VonMises],
    dist2: Union[VonMises, tfd.VonMises],
    *unused_args, **unused_kwargs,
) -> Array:
  """Batched KL divergence KL(d1 || d2) between von Mises distributions.

  Args:
    dist1: A VonMises distribution.
    dist2: A VonMises distribution.

  Returns:
    Batchwise `KL(dist1 || dist2)`.
  """
  i0e_concentration1 = jax.scipy.special.i0e(dist1.concentration)
  i1e_concentration1 = jax.scipy.special.i1e(dist1.concentration)
  i0e_concentration2 = jax.scipy.special.i0e(dist2.concentration)
  return (
      (dist2.concentration - dist1.concentration) +
      jnp.log(i0e_concentration2 / i0e_concentration1) +
      (i1e_concentration1 / i0e_concentration1) * (
          dist1.concentration
          - dist2.concentration * jnp.cos(dist1.loc - dist2.loc)
      )
  )


# Register the KL functions with TFP.
tfd.RegisterKL(VonMises, VonMises)(_kl_divergence_vonmises_vonmises)
tfd.RegisterKL(VonMises, VonMises.equiv_tfp_cls)(
    _kl_divergence_vonmises_vonmises
)
tfd.RegisterKL(VonMises.equiv_tfp_cls, VonMises)(
    _kl_divergence_vonmises_vonmises
)



# Copyright 2021 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""OneHotCategorical distribution."""

from typing import Any, Optional, Tuple, Union

import chex
from distrax._src.distributions import categorical
from distrax._src.distributions import distribution
from distrax._src.utils import math
import jax
import jax.numpy as jnp
from tensorflow_probability.substrates import jax as tfp


tfd = tfp.distributions

Array = chex.Array
PRNGKey = chex.PRNGKey
EventT = distribution.EventT


class OneHotCategorical(categorical.Categorical):
  """OneHotCategorical distribution."""

  equiv_tfp_cls = tfd.OneHotCategorical

  def __init__(self,
               logits: Optional[Array] = None,
               probs: Optional[Array] = None,
               dtype: Union[jnp.dtype, type[Any]] = int):
    """Initializes a OneHotCategorical distribution.

    Args:
      logits: Logit transform of the probability of each category. Only one
        of `logits` or `probs` can be specified.
      probs: Probability of each category. Only one of `logits` or `probs` can
        be specified.
      dtype: The type of event samples.
    """
    super().__init__(logits=logits, probs=probs, dtype=dtype)

  @property
  def event_shape(self) -> Tuple[int, ...]:
    """Shape of event of distribution samples."""
    return (self.num_categories,)

  def _sample_n(self, key: PRNGKey, n: int) -> Array:
    """See `Distribution._sample_n`."""
    new_shape = (n,) + self.logits.shape[:-1]
    is_valid = jnp.logical_and(
        jnp.all(jnp.isfinite(self.probs), axis=-1, keepdims=True),
        jnp.all(self.probs >= 0, axis=-1, keepdims=True))
    draws = jax.random.categorical(
        key=key, logits=self.logits, axis=-1, shape=new_shape)
    draws_one_hot = jax.nn.one_hot(
        draws, num_classes=self.num_categories).astype(self._dtype)
    return jnp.where(is_valid, draws_one_hot, jnp.ones_like(draws_one_hot) * -1)

  def log_prob(self, value: EventT) -> Array:
    """See `Distribution.log_prob`."""
    return jnp.sum(math.multiply_no_nan(self.logits, value), axis=-1)

  def prob(self, value: EventT) -> Array:
    """See `Distribution.prob`."""
    return jnp.sum(math.multiply_no_nan(self.probs, value), axis=-1)

  def mode(self) -> Array:
    """Calculates the mode."""
    preferences = self._probs if self._logits is None else self._logits
    assert preferences is not None
    greedy_index = jnp.argmax(preferences, axis=-1)
    return jax.nn.one_hot(greedy_index, self.num_categories).astype(self._dtype)

  def cdf(self, value: EventT) -> Array:
    """See `Distribution.cdf`."""
    return jnp.sum(math.multiply_no_nan(
        jnp.cumsum(self.probs, axis=-1), value), axis=-1)

  def __getitem__(self, index) -> 'OneHotCategorical':
    """See `Distribution.__getitem__`."""
    index = distribution.to_batch_shape_index(self.batch_shape, index)
    if self._logits is not None:
      return OneHotCategorical(logits=self.logits[index], dtype=self._dtype)
    return OneHotCategorical(probs=self.probs[index], dtype=self._dtype)



# Copyright 2021 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Split coupling bijector."""

from typing import Any, Callable, Tuple

from distrax._src.bijectors import bijector as base
from distrax._src.bijectors import block
from distrax._src.utils import conversion
import jax.numpy as jnp


Array = base.Array
BijectorParams = Any


class SplitCoupling(base.Bijector):
  """Split coupling bijector, with arbitrary conditioner & inner bijector.

  This coupling bijector splits the input array into two parts along a specified
  axis. One part remains unchanged, whereas the other part is transformed by an
  inner bijector conditioned on the unchanged part.

  Let `f` be a conditional bijector (the inner bijector) and `g` be a function
  (the conditioner). For `swap=False`, the split coupling bijector is defined as
  follows:

  - Forward:
    ```
    x = [x1, x2]
    y1 = x1
    y2 = f(x2; g(x1))
    y = [y1, y2]
    ```

  - Forward Jacobian log determinant:
    ```
    x = [x1, x2]
    log|det J(x)| = log|det df/dx2(x2; g(x1))|
    ```

  - Inverse:
    ```
    y = [y1, y2]
    x1 = y1
    x2 = f^{-1}(y2; g(y1))
    x = [x1, x2]
    ```

  - Inverse Jacobian log determinant:
    ```
    y = [y1, y2]
    log|det J(y)| = log|det df^{-1}/dy2(y2; g(y1))|
    ```

  Here, `[x1, x2]` is a partition of `x` along some axis. By default, `x1`
  remains unchanged and `x2` is transformed. If `swap=True`, `x2` will remain
  unchanged and `x1` will be transformed.
  """

  def __init__(self,
               split_index: int,
               event_ndims: int,
               conditioner: Callable[[Array], BijectorParams],
               bijector: Callable[[BijectorParams], base.BijectorLike],
               swap: bool = False,
               split_axis: int = -1):
    """Initializes a SplitCoupling bijector.

    Args:
      split_index: the index used to split the input. The input array will be
        split along the axis specified by `split_axis` into two parts. The first
        part will correspond to indices up to `split_index` (non-inclusive),
        whereas the second part will correspond to indices starting from
        `split_index` (inclusive).
      event_ndims: the number of event dimensions the bijector operates on. The
        `event_ndims_in` and `event_ndims_out` of the coupling bijector are both
        equal to `event_ndims`.
      conditioner: a function that computes the parameters of the inner bijector
        as a function of the unchanged part of the input. The output of the
        conditioner will be passed to `bijector` in order to obtain the inner
        bijector.
      bijector: a callable that returns the inner bijector that will be used to
        transform one of the two parts. The input to `bijector` is a set of
        parameters that can be used to configure the inner bijector. The
        `event_ndims_in` and `event_ndims_out` of the inner bijector must be
        equal, and less than or equal to `event_ndims`. If they are less than
        `event_ndims`, the remaining dimensions will be converted to event
        dimensions using `distrax.Block`.
      swap: by default, the part of the input up to `split_index` is the one
        that remains unchanged. If `swap` is True, then the other part remains
        unchanged and the first one is transformed instead.
      split_axis: the axis along which to split the input. Must be negative,
        that is, it must index from the end. By default, it's the last axis.
    """
    if split_index < 0:
      raise ValueError(
          f'The split index must be non-negative; got {split_index}.')
    if split_axis >= 0:
      raise ValueError(f'The split axis must be negative; got {split_axis}.')
    if event_ndims < 0:
      raise ValueError(
          f'`event_ndims` must be non-negative; got {event_ndims}.')
    if split_axis < -event_ndims:
      raise ValueError(
          f'The split axis points to an axis outside the event. With '
          f'`event_ndims == {event_ndims}`, the split axis must be between -1 '
          f'and {-event_ndims}. Got `split_axis == {split_axis}`.')
    self._split_index = split_index
    self._conditioner = conditioner
    self._bijector = bijector
    self._swap = swap
    self._split_axis = split_axis
    super().__init__(event_ndims_in=event_ndims)

  @property
  def bijector(self) -> Callable[[BijectorParams], base.BijectorLike]:
    """The callable that returns the inner bijector of `SplitCoupling`."""
    return self._bijector

  @property
  def conditioner(self) -> Callable[[Array], BijectorParams]:
    """The conditioner function."""
    return self._conditioner

  @property
  def split_index(self) -> int:
    """The index used to split the input."""
    return self._split_index

  @property
  def swap(self) -> bool:
    """The flag that determines which part of the input remains unchanged."""
    return self._swap

  @property
  def split_axis(self) -> int:
    """The axis along which to split the input."""
    return self._split_axis

  def _split(self, x: Array) -> Tuple[Array, Array]:
    x1, x2 = jnp.split(x, [self._split_index], self._split_axis)
    if self._swap:
      x1, x2 = x2, x1
    return x1, x2

  def _recombine(self, x1: Array, x2: Array) -> Array:
    if self._swap:
      x1, x2 = x2, x1
    return jnp.concatenate([x1, x2], self._split_axis)

  def _inner_bijector(self, params: BijectorParams) -> base.Bijector:
    """Returns an inner bijector for the passed params."""
    bijector = conversion.as_bijector(self._bijector(params))
    if bijector.event_ndims_in != bijector.event_ndims_out:
      raise ValueError(
          f'The inner bijector must have `event_ndims_in==event_ndims_out`. '
          f'Instead, it has `event_ndims_in=={bijector.event_ndims_in}` and '
          f'`event_ndims_out=={bijector.event_ndims_out}`.')
    extra_ndims = self.event_ndims_in - bijector.event_ndims_in
    if extra_ndims < 0:
      raise ValueError(
          f'The inner bijector can\'t have more event dimensions than the '
          f'coupling bijector. Got {bijector.event_ndims_in} for the inner '
          f'bijector and {self.event_ndims_in} for the coupling bijector.')
    elif extra_ndims > 0:
      bijector = block.Block(bijector, extra_ndims)
    return bijector

  def forward_and_log_det(self, x: Array) -> Tuple[Array, Array]:
    """Computes y = f(x) and log|det J(f)(x)|."""
    self._check_forward_input_shape(x)
    x1, x2 = self._split(x)
    params = self._conditioner(x1)
    y2, logdet = self._inner_bijector(params).forward_and_log_det(x2)
    return self._recombine(x1, y2), logdet

  def inverse_and_log_det(self, y: Array) -> Tuple[Array, Array]:
    """Computes x = f^{-1}(y) and log|det J(f^{-1})(y)|."""
    self._check_inverse_input_shape(y)
    y1, y2 = self._split(y)
    params = self._conditioner(y1)
    x2, logdet = self._inner_bijector(params).inverse_and_log_det(y2)
    return self._recombine(y1, x2), logdet



# Copyright 2021 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Tests for `greedy.py`."""

from absl.testing import absltest
from absl.testing import parameterized

import chex
from distrax._src.distributions import greedy
from distrax._src.utils import equivalence
import jax.experimental
import jax.numpy as jnp
import numpy as np


class GreedyTest(equivalence.EquivalenceTest):

  def setUp(self):
    super().setUp()
    self._init_distr_cls(greedy.Greedy)
    self.preferences = jnp.array([0., 4., -1., 4.])

  def test_parameters_from_preferences(self):
    dist = self.distrax_cls(preferences=self.preferences)
    expected_probs = jnp.array([0., 0.5, 0., 0.5])
    self.assertion_fn(rtol=2e-3)(dist.logits, jnp.log(expected_probs))
    self.assertion_fn(rtol=2e-3)(dist.probs, expected_probs)

  def test_num_categories(self):
    dist = self.distrax_cls(preferences=self.preferences)
    np.testing.assert_equal(dist.num_categories, len(self.preferences))

  @chex.all_variants
  @parameterized.named_parameters(
      ('int32', jnp.int32),
      ('int64', jnp.int64),
      ('float32', jnp.float32),
      ('float64', jnp.float64))
  def test_sample_dtype(self, dtype):
    with jax.experimental.enable_x64(dtype.dtype.itemsize == 8):
      dist = self.distrax_cls(preferences=self.preferences, dtype=dtype)
      samples = self.variant(dist.sample)(seed=self.key)
      self.assertEqual(samples.dtype, dist.dtype)
      chex.assert_type(samples, dtype)

  def test_jittable(self):
    super()._test_jittable((np.array([0., 4., -1., 4.]),))

  @parameterized.named_parameters(
      ('single element', 2),
      ('range', slice(-1)),
      ('range_2', (slice(None), slice(-1))),
  )
  def test_slice(self, slice_):
    preferences = np.abs(np.random.randn(3, 4, 5))
    dtype = jnp.float32
    dist = self.distrax_cls(preferences, dtype=dtype)
    dist_sliced = dist[slice_]
    self.assertIsInstance(dist_sliced, greedy.Greedy)
    self.assertion_fn(rtol=2e-3)(dist_sliced.preferences, preferences[slice_])
    self.assertEqual(dist_sliced.dtype, dtype)

  def test_slice_ellipsis(self):
    preferences = np.abs(np.random.randn(3, 4, 5))
    dtype = jnp.float32
    dist = self.distrax_cls(preferences, dtype=dtype)
    dist_sliced = dist[..., -1]
    self.assertIsInstance(dist_sliced, greedy.Greedy)
    self.assertion_fn(rtol=2e-3)(dist_sliced.preferences, preferences[:, -1])
    self.assertEqual(dist_sliced.dtype, dtype)


if __name__ == '__main__':
  absltest.main()



# Copyright 2021 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Tests for `categorical.py`."""

from absl.testing import absltest
from absl.testing import parameterized

import chex
from distrax._src.distributions import categorical
from distrax._src.utils import equivalence
from distrax._src.utils import math
import jax
import jax.numpy as jnp
import numpy as np


class CategoricalTest(equivalence.EquivalenceTest):

  def setUp(self):
    super().setUp()
    self._init_distr_cls(categorical.Categorical)

  @parameterized.named_parameters(
      ('1d probs', (4,), True),
      ('1d logits', (4,), False),
      ('2d probs', (3, 4), True),
      ('2d logits', (3, 4), False),
  )
  def test_properties(self, shape, from_probs):
    rng = np.random.default_rng(42)
    probs = rng.uniform(size=shape)  # Intentional unnormalization of `probs`.
    logits = np.log(probs)
    dist_kwargs = {'probs': probs} if from_probs else {'logits': logits}
    dist = self.distrax_cls(**dist_kwargs)
    self.assertEqual(dist.event_shape, ())
    self.assertEqual(dist.batch_shape, shape[:-1])
    self.assertEqual(dist.num_categories, shape[-1])
    self.assertion_fn(rtol=1e-3)(dist.logits, math.normalize(logits=logits))
    self.assertion_fn(rtol=1e-3)(dist.probs, math.normalize(probs=probs))

  @parameterized.named_parameters(
      ('probs and logits', {'logits': [0.1, -0.2], 'probs': [0.6, 0.4]}),
      ('both probs and logits are None', {'logits': None, 'probs': None}),
      ('bool dtype', {'logits': [0.1, -0.2], 'dtype': jnp.bool_}),
      ('complex64 dtype', {'logits': [0.1, -0.2], 'dtype': jnp.complex64}),
      ('complex128 dtype', {'logits': [0.1, -0.2], 'dtype': jnp.complex128}),
  )
  def test_raises_on_invalid_inputs(self, dist_params):
    with self.assertRaises(ValueError):
      self.distrax_cls(**dist_params)

  @chex.all_variants
  @parameterized.named_parameters(
      ('1d logits, no shape', {'logits': [0.0, 1.0, -0.5]}, ()),
      ('1d probs, no shape', {'probs': [0.2, 0.5, 0.3]}, ()),
      ('1d logits, int shape', {'logits': [0.0, 1.0, -0.5]}, 1),
      ('1d probs, int shape', {'probs': [0.2, 0.5, 0.3]}, 1),
      ('1d logits, 1-tuple shape', {'logits': [0.0, 1.0, -0.5]}, (1,)),
      ('1d probs, 1-tuple shape', {'probs': [0.2, 0.5, 0.3]}, (1,)),
      ('1d logits, 2-tuple shape', {'logits': [0.0, 1.0, -0.5]}, (5, 4)),
      ('1d probs, 2-tuple shape', {'probs': [0.2, 0.5, 0.3]}, (5, 4)),
      ('2d logits, no shape', {'logits': [[0.0, 1.0, -0.5], [-0.1, 0.3, 0.0]]},
       ()),
      ('2d probs, no shape', {'probs': [[0.1, 0.4, 0.5], [0.5, 0.25, 0.25]]},
       ()),
      ('2d logits, int shape', {'logits': [[0.0, 1.0, -0.5], [-0.1, 0.3, 0.0]]},
       4),
      ('2d probs, int shape', {'probs': [[0.1, 0.4, 0.5], [0.5, 0.25, 0.25]]},
       4),
      ('2d logits, 1-tuple shape',
       {'logits': [[0.0, 1.0, -0.5], [-0.1, 0.3, 0.0]]}, (5,)),
      ('2d probs, 1-tuple shape',
       {'probs': [[0.1, 0.4, 0.5], [0.5, 0.25, 0.25]]}, (5,)),
      ('2d logits, 2-tuple shape',
       {'logits': [[0.0, 1.0, -0.5], [-0.1, 0.3, 0.0]]}, (5, 4)),
      ('2d probs, 2-tuple shape',
       {'probs': [[0.1, 0.4, 0.5], [0.5, 0.25, 0.25]]}, (5, 4)),
  )
  def test_sample_shape(self, distr_params, sample_shape):
    distr_params = {k: jnp.asarray(v) for k, v in distr_params.items()}
    super()._test_sample_shape(
        dist_args=(),
        dist_kwargs=distr_params,
        sample_shape=sample_shape)

  @chex.all_variants
  @parameterized.named_parameters(
      ('1d logits, no shape', {'logits': [0.0, 1.0, -0.5]}, ()),
      ('1d probs, no shape', {'probs': [0.2, 0.5, 0.3]}, ()),
      ('1d logits, int shape', {'logits': [0.0, 1.0, -0.5]}, 1),
      ('1d probs, int shape', {'probs': [0.2, 0.5, 0.3]}, 1),
      ('1d logits, 1-tuple shape', {'logits': [0.0, 1.0, -0.5]}, (1,)),
      ('1d probs, 1-tuple shape', {'probs': [0.2, 0.5, 0.3]}, (1,)),
      ('1d logits, 2-tuple shape', {'logits': [0.0, 1.0, -0.5]}, (5, 4)),
      ('1d probs, 2-tuple shape', {'probs': [0.2, 0.5, 0.3]}, (5, 4)),
      ('2d logits, no shape', {'logits': [[0.0, 1.0, -0.5], [-0.1, 0.3, 0.0]]},
       ()),
      ('2d probs, no shape', {'probs': [[0.1, 0.4, 0.5], [0.5, 0.25, 0.25]]},
       ()),
      ('2d logits, int shape', {'logits': [[0.0, 1.0, -0.5], [-0.1, 0.3, 0.0]]},
       4),
      ('2d probs, int shape', {'probs': [[0.1, 0.4, 0.5], [0.5, 0.25, 0.25]]},
       4),
      ('2d logits, 1-tuple shape',
       {'logits': [[0.0, 1.0, -0.5], [-0.1, 0.3, 0.0]]}, (5,)),
      ('2d probs, 1-tuple shape',
       {'probs': [[0.1, 0.4, 0.5], [0.5, 0.25, 0.25]]}, (5,)),
      ('2d logits, 2-tuple shape',
       {'logits': [[0.0, 1.0, -0.5], [-0.1, 0.3, 0.0]]}, (5, 4)),
      ('2d probs, 2-tuple shape',
       {'probs': [[0.1, 0.4, 0.5], [0.5, 0.25, 0.25]]}, (5, 4)),
  )
  def test_sample_and_log_prob(self, distr_params, sample_shape):
    distr_params = {k: jnp.asarray(v) for k, v in distr_params.items()}
    super()._test_sample_and_log_prob(
        dist_args=(),
        dist_kwargs=distr_params,
        sample_shape=sample_shape,
        assertion_fn=self.assertion_fn(rtol=2e-3))

  @chex.all_variants
  @parameterized.named_parameters(
      ('sample, int16', 'sample', jnp.int16),
      ('sample, int32', 'sample', jnp.int32),
      ('sample, uint16', 'sample', jnp.uint16),
      ('sample, uint32', 'sample', jnp.uint32),
      ('sample, float16', 'sample', jnp.float16),
      ('sample, float32', 'sample', jnp.float32),
      ('sample_and_log_prob, int16', 'sample_and_log_prob', jnp.int16),
      ('sample_and_log_prob, int32', 'sample_and_log_prob', jnp.int32),
      ('sample_and_log_prob, uint16', 'sample_and_log_prob', jnp.uint16),
      ('sample_and_log_prob, uint32', 'sample_and_log_prob', jnp.uint32),
      ('sample_and_log_prob, float16', 'sample_and_log_prob', jnp.float16),
      ('sample_and_log_prob, float32', 'sample_and_log_prob', jnp.float32),
  )
  def test_sample_dtype(self, method, dtype):
    dist_params = {'logits': [0.1, -0.1, 0.5, -0.8, 1.5], 'dtype': dtype}
    dist = self.distrax_cls(**dist_params)
    samples = self.variant(getattr(dist, method))(seed=self.key)
    samples = samples[0] if method == 'sample_and_log_prob' else samples
    self.assertEqual(samples.dtype, dist.dtype)
    self.assertEqual(samples.dtype, dtype)

  @chex.all_variants
  @parameterized.named_parameters(
      ('sample, from probs', 'sample', True),
      ('sample, from logits', 'sample', False),
      ('sample_and_log_prob, from probs', 'sample_and_log_prob', True),
      ('sample_and_log_prob, from logits', 'sample_and_log_prob', False),
  )
  def test_sample_values(self, method, from_probs):
    probs = np.array([[0.5, 0.25, 0.25], [0., 0., 1.]])  # Includes edge case.
    num_categories = probs.shape[-1]
    logits = np.log(probs)
    n_samples = 100000
    dist_kwargs = {'probs': probs} if from_probs else {'logits': logits}
    dist = self.distrax_cls(**dist_kwargs)
    sample_fn = self.variant(
        lambda key: getattr(dist, method)(seed=key, sample_shape=n_samples))
    samples = sample_fn(self.key)
    samples = samples[0] if method == 'sample_and_log_prob' else samples
    self.assertEqual(samples.shape, (n_samples,) + probs.shape[:-1])
    self.assertTrue(np.all(
        np.logical_and(samples >= 0, samples < num_categories)))
    np.testing.assert_array_equal(jnp.floor(samples), samples)
    samples_one_hot = jax.nn.one_hot(samples, num_categories, axis=-1)
    self.assertion_fn(rtol=0.1)(np.mean(samples_one_hot, axis=0), probs)

  @chex.all_variants
  @parameterized.named_parameters(
      ('sample', 'sample'),
      ('sample_and_log_prob', 'sample_and_log_prob'),
  )
  def test_sample_values_invalid_probs(self, method):
    # Check that samples=-1 if probs are negative or NaN after normalization.
    n_samples = 1000
    probs = np.array([
        [0.1, -0.4, 0.2, 0.3],  # Negative probabilities.
        [-0.1, 0.1, 0.0, 0.0],  # NaN probabilities after normalization.
        [0.1, 0.25, 0.2, 0.8],  # Valid (unnormalized) probabilities.
    ])
    dist = self.distrax_cls(probs=probs)
    sample_fn = self.variant(
        lambda key: getattr(dist, method)(seed=key, sample_shape=n_samples))
    samples = sample_fn(self.key)
    samples = samples[0] if method == 'sample_and_log_prob' else samples
    self.assertion_fn(rtol=1e-4)(samples[..., :-1], -1)
    np.testing.assert_array_compare(lambda x, y: x >= y, samples[..., -1], 0)

  @chex.all_variants
  @parameterized.named_parameters(
      ('1d logits, 1d value', {'logits': [0.0, 0.5, -0.5]}, [1, 0, 2, 0]),
      ('1d probs, 1d value', {'probs': [0.3, 0.2, 0.5]}, [1, 0, 2, 0]),
      ('1d logits, 2d value', {'logits': [0.0, 0.5, -0.5]}, [[1, 0], [2, 0]]),
      ('1d probs, 2d value', {'probs': [0.3, 0.2, 0.5]}, [[1, 0], [2, 0]]),
      ('2d logits, 1d value', {'logits': [[0.0, 0.5, -0.5], [-0.2, 0.3, 0.5]]},
       [1, 0]),
      ('2d probs, 1d value', {'probs': [[0.1, 0.5, 0.4], [0.3, 0.3, 0.4]]},
       [1, 0]),
      ('2d logits, 2d value', {'logits': [[0.0, 0.5, -0.5], [-0.2, 0.3, 0.5]]},
       [[1, 0], [2, 1]]),
      ('2d probs, 2d value', {'probs': [[0.1, 0.5, 0.4], [0.3, 0.3, 0.4]]},
       [[1, 0], [2, 1]]),
      ('extreme probs', {'probs': [0.0, 1.0, 0.0]}, [0, 1, 1, 2]),
  )
  def test_method_with_input(self, distr_params, value):
    distr_params = {k: jnp.asarray(v) for k, v in distr_params.items()}
    value = np.asarray(value, dtype=np.int32)
    for method in ['prob', 'log_prob', 'cdf', 'log_cdf', 'survival_function']:
      with self.subTest(method=method):
        super()._test_attribute(
            attribute_string=method,
            dist_kwargs=distr_params,
            call_args=(value,),
            assertion_fn=self.assertion_fn(atol=3e-5))
    # We separate the `log_survival_function` method because TFP sometimes gives
    # NaN instead of `-jnp.inf` when evaluated at `num_categories - 1` on a TPU
    # if the distribution was created using the logits parameter.
    dist = self.distrax_cls(**distr_params)
    tfp_dist = self.distrax_cls.equiv_tfp_cls(**distr_params)
    num_categories = dist.num_categories
    log_sf = tfp_dist.log_survival_function(value)
    log_sf = jnp.where(value == num_categories - 1, -jnp.inf, log_sf)
    with self.subTest(method='log_survival_function'):
      self.assertion_fn(atol=3e-5)(
          self.variant(dist.log_survival_function)(value), log_sf)

  @chex.all_variants
  def test_method_with_input_unnormalized_probs(self):
    # We test this case separately because the result of `cdf` and `log_cdf`
    # differs from TFP when the input `probs` are not normalized.
    probs = np.array([0.1, 0.2, 0.3])
    normalized_probs = probs / np.sum(probs, axis=-1, keepdims=True)
    distr_params = {'probs': probs}
    value = np.asarray([0, 1, 2], dtype=np.int32)
    dist = self.distrax_cls(**distr_params)
    self.assertion_fn(rtol=1e-3)(
        self.variant(dist.prob)(value), normalized_probs)
    self.assertion_fn(rtol=1e-3)(
        self.variant(dist.log_prob)(value), np.log(normalized_probs))
    self.assertion_fn(rtol=1e-3)(
        self.variant(dist.cdf)(value), np.cumsum(normalized_probs))
    self.assertion_fn(atol=5e-5)(
        self.variant(dist.log_cdf)(value), np.log(np.cumsum(normalized_probs)))
    self.assertion_fn(atol=1e-5)(
        self.variant(dist.survival_function)(value),
        1. - np.cumsum(normalized_probs))
    # In the line below, we compare against `jnp` instead of `np` because the
    # latter gives `1. - np.cumsum(normalized_probs)[-1] = 1.1e-16` instead of
    # `0.`, so its log is innacurate: it gives `-36.7` instead of `-np.inf`.
    self.assertion_fn(atol=1e-5)(
        self.variant(dist.log_survival_function)(value),
        jnp.log(1. - jnp.cumsum(normalized_probs)))

  @chex.all_variants
  def test_method_with_input_outside_domain(self):
    probs = jnp.asarray([0.2, 0.3, 0.5])
    dist = self.distrax_cls(probs=probs)
    value = jnp.asarray([-1, -2, 3, 4], dtype=jnp.int32)
    self.assertion_fn(atol=1e-5)(
        self.variant(dist.prob)(value), np.asarray([0., 0., 0., 0.]))
    self.assertTrue(np.all(self.variant(dist.log_prob)(value) == -jnp.inf))
    self.assertion_fn(atol=1e-5)(
        self.variant(dist.cdf)(value), np.asarray([0., 0., 1., 1.]))
    self.assertion_fn(rtol=1e-3)(
        self.variant(dist.log_cdf)(value), np.log(np.asarray([0., 0., 1., 1.])))
    self.assertion_fn(atol=1e-5)(
        self.variant(dist.survival_function)(value),
        np.asarray([1., 1., 0., 0.]))
    self.assertion_fn(atol=1e-5)(
        self.variant(dist.log_survival_function)(value),
        np.log(np.asarray([1., 1., 0., 0.])))

  @chex.all_variants(with_pmap=False)
  @parameterized.named_parameters(
      ('2d probs', True),
      ('2d logits', False),
  )
  def test_method(self, from_probs):
    rng = np.random.default_rng(42)
    probs = rng.uniform(size=(4, 3))
    probs /= np.sum(probs, axis=-1, keepdims=True)
    logits = np.log(probs)
    distr_params = {'probs': probs} if from_probs else {'logits': logits}
    for method in ['entropy', 'mode', 'logits_parameter']:
      with self.subTest(method=method):
        super()._test_attribute(
            attribute_string=method,
            dist_kwargs=distr_params,
            call_args=(),
            assertion_fn=self.assertion_fn(rtol=1e-3))

  @chex.all_variants(with_pmap=False)
  @parameterized.named_parameters(
      ('kl distrax_to_distrax', 'kl_divergence', 'distrax_to_distrax'),
      ('kl distrax_to_tfp', 'kl_divergence', 'distrax_to_tfp'),
      ('kl tfp_to_distrax', 'kl_divergence', 'tfp_to_distrax'),
      ('cross-ent distrax_to_distrax', 'cross_entropy', 'distrax_to_distrax'),
      ('cross-ent distrax_to_tfp', 'cross_entropy', 'distrax_to_tfp'),
      ('cross-ent tfp_to_distrax', 'cross_entropy', 'tfp_to_distrax'))
  def test_with_two_distributions(self, function_string, mode_string):
    super()._test_with_two_distributions(
        attribute_string=function_string,
        mode_string=mode_string,
        dist1_kwargs={
            'probs':
                jnp.asarray([[0.1, 0.5, 0.4], [0.2, 0.4, 0.4]])
        },
        dist2_kwargs={
            'logits': jnp.asarray([0.0, 0.1, 0.1]),
        },
        assertion_fn=self.assertion_fn(rtol=2e-3))

  @chex.all_variants(with_pmap=False)
  @parameterized.named_parameters(
      ('kl distrax_to_distrax', 'kl_divergence', 'distrax_to_distrax'),
      ('kl distrax_to_tfp', 'kl_divergence', 'distrax_to_tfp'),
      ('kl tfp_to_distrax', 'kl_divergence', 'tfp_to_distrax'),
      ('cross-ent distrax_to_distrax', 'cross_entropy', 'distrax_to_distrax'),
      ('cross-ent distrax_to_tfp', 'cross_entropy', 'distrax_to_tfp'),
      ('cross-ent tfp_to_distrax', 'cross_entropy', 'tfp_to_distrax'),
  )
  def test_with_two_distributions_extreme_cases(
      self, function_string, mode_string):
    super()._test_with_two_distributions(
        attribute_string=function_string,
        mode_string=mode_string,
        dist1_kwargs={
            'probs':
                jnp.asarray([[0.1, 0.5, 0.4], [0.4, 0.0, 0.6], [0.4, 0.6, 0.]])
        },
        dist2_kwargs={
            'logits': jnp.asarray([0.0, 0.1, -jnp.inf]),
        },
        assertion_fn=self.assertion_fn(atol=5e-5))

  @chex.all_variants(with_pmap=False)
  @parameterized.named_parameters(
      ('kl distrax_to_distrax', 'kl_divergence', 'distrax_to_distrax'),
      ('kl distrax_to_tfp', 'kl_divergence', 'distrax_to_tfp'),
      ('kl tfp_to_distrax', 'kl_divergence', 'tfp_to_distrax'),
      ('cross-ent distrax_to_distrax', 'cross_entropy', 'distrax_to_distrax'),
      ('cross-ent distrax_to_tfp', 'cross_entropy', 'distrax_to_tfp'),
      ('cross-ent tfp_to_distrax', 'cross_entropy', 'tfp_to_distrax'),
  )
  def test_with_two_distributions_raises_on_invalid_num_categories(
      self, function_string, mode_string):
    probs1 = jnp.asarray([0.1, 0.5, 0.4])
    distrax_dist1 = self.distrax_cls(probs=probs1)
    tfp_dist1 = self.distrax_cls.equiv_tfp_cls(probs=probs1)
    logits2 = jnp.asarray([-0.1, 0.3])
    distrax_dist2 = self.distrax_cls(logits=logits2)
    tfp_dist2 = self.distrax_cls.equiv_tfp_cls(logits=logits2)
    dist_a = tfp_dist1 if mode_string == 'tfp_to_distrax' else distrax_dist1
    dist_b = tfp_dist2 if mode_string == 'distrax_to_tfp' else distrax_dist2
    first_fn = self.variant(getattr(dist_a, function_string))
    with self.assertRaises(ValueError):
      _ = first_fn(dist_b)
    dist_a = tfp_dist2 if mode_string == 'tfp_to_distrax' else distrax_dist2
    dist_b = tfp_dist1 if mode_string == 'distrax_to_tfp' else distrax_dist1
    second_fn = self.variant(getattr(dist_a, function_string))
    with self.assertRaises(ValueError):
      _ = second_fn(dist_b)

  def test_jittable(self):
    super()._test_jittable((np.array([0., 4., -1., 4.]),))

  @parameterized.named_parameters(
      ('single element, from probs', 2, True),
      ('single element, from logits', 2, False),
      ('range, from probs', slice(-1), True),
      ('range, from logits', slice(-1), False),
      ('range_2, from probs', (slice(None), slice(-1)), True),
      ('range_2, from logits', (slice(None), slice(-1)), False),
  )
  def test_slice(self, slice_, from_probs):
    rng = np.random.default_rng(42)
    logits = rng.normal(size=(3, 4, 5))
    probs = jax.nn.softmax(logits, axis=-1)
    dist_kwargs = {'probs': probs} if from_probs else {'logits': logits}
    dist = self.distrax_cls(**dist_kwargs)
    self.assertion_fn(rtol=1e-3)(
        dist[slice_].logits, math.normalize(logits=logits[slice_]))
    self.assertion_fn(rtol=1e-3)(dist[slice_].probs, probs[slice_])

  @parameterized.named_parameters(
      ('from probs', True),
      ('from logits', False),
  )
  def test_slice_ellipsis(self, from_probs):
    rng = np.random.default_rng(42)
    logits = rng.normal(size=(3, 4, 5))
    probs = jax.nn.softmax(logits, axis=-1)
    dist_kwargs = {'probs': probs} if from_probs else {'logits': logits}
    dist = self.distrax_cls(**dist_kwargs)
    self.assertion_fn(rtol=1e-3)(
        dist[..., -1].logits, math.normalize(logits=logits[:, -1, :]))
    self.assertion_fn(rtol=1e-3)(dist[..., -1].probs, probs[:, -1, :])

  def test_vmap_inputs(self):
    def log_prob_sum(dist, x):
      return dist.log_prob(x).sum()

    dist = categorical.Categorical(jnp.arange(3 * 4 * 5).reshape((3, 4, 5)))
    x = jnp.zeros((3, 4), jnp.int_)

    with self.subTest('no vmap'):
      actual = log_prob_sum(dist, x)
      expected = dist.log_prob(x).sum()
      self.assertion_fn()(actual, expected)

    with self.subTest('axis=0'):
      actual = jax.vmap(log_prob_sum, in_axes=0)(dist, x)
      expected = dist.log_prob(x).sum(axis=1)
      self.assertion_fn()(actual, expected)

    with self.subTest('axis=1'):
      actual = jax.vmap(log_prob_sum, in_axes=1)(dist, x)
      expected = dist.log_prob(x).sum(axis=0)
      self.assertion_fn()(actual, expected)

  def test_vmap_outputs(self):
    def summed_dist(logits):
      logits1 = logits.sum(keepdims=True)
      logits2 = -logits1
      logits = jnp.concatenate([logits1, logits2], axis=-1)
      return categorical.Categorical(logits)

    logits = jnp.arange((3 * 4 * 5)).reshape((3, 4, 5))
    actual = jax.vmap(summed_dist)(logits)

    logits1 = logits.sum(axis=(1, 2), keepdims=True)
    logits2 = -logits1
    logits = jnp.concatenate([logits1, logits2], axis=-1)
    expected = categorical.Categorical(logits)

    np.testing.assert_equal(actual.batch_shape, expected.batch_shape)
    np.testing.assert_equal(actual.event_shape, expected.event_shape)

    x = jnp.array([[[0]], [[1]], [[0]]], jnp.int_)
    self.assertion_fn(rtol=1e-6)(actual.log_prob(x), expected.log_prob(x))

  @parameterized.named_parameters(
      ('-inf logits', np.array([-jnp.inf, 2, -3, -jnp.inf, 5.0])),
      ('uniform large negative logits', np.array([-1e9] * 11)),
      ('uniform large positive logits', np.array([1e9] * 11)),
      ('uniform', np.array([0.0] * 11)),
      ('typical', np.array([1, 7, -3, 2, 4.0])),
  )
  def test_entropy_grad(self, logits):
    clipped_logits = jnp.maximum(-10000, logits)

    def entropy_fn(logits):
      return categorical.Categorical(logits).entropy()
    entropy, grads = jax.value_and_grad(entropy_fn)(logits)
    expected_entropy, expected_grads = jax.value_and_grad(entropy_fn)(
        clipped_logits)
    self.assertion_fn(rtol=1e-6)(expected_entropy, entropy)
    self.assertion_fn(rtol=1e-6)(expected_grads, grads)
    self.assertTrue(np.isfinite(entropy).all())
    self.assertTrue(np.isfinite(grads).all())

  @parameterized.named_parameters(
      ('-inf logits1', np.array([-jnp.inf, 2, -3, -jnp.inf, 5.0]),
       np.array([1, 7, -3, 2, 4.0])),
      ('-inf logits both', np.array([-jnp.inf, 2, -1000, -jnp.inf, 5.0]),
       np.array([-jnp.inf, 7, -jnp.inf, 2, 4.0])),
      ('typical', np.array([5, -2, 0, 1, 4.0]),
       np.array([1, 7, -3, 2, 4.0])),
  )
  def test_kl_grad(self, logits1, logits2):
    clipped_logits1 = jnp.maximum(-10000, logits1)
    clipped_logits2 = jnp.maximum(-10000, logits2)

    def kl_fn(logits1, logits2):
      return categorical.Categorical(logits1).kl_divergence(
          categorical.Categorical(logits2))
    kl, grads = jax.value_and_grad(
        kl_fn, argnums=(0, 1))(logits1, logits2)
    expected_kl, expected_grads = jax.value_and_grad(
        kl_fn, argnums=(0, 1))(clipped_logits1, clipped_logits2)
    self.assertion_fn(rtol=1e-6)(expected_kl, kl)
    self.assertion_fn(rtol=1e-6)(expected_grads, grads)
    self.assertTrue(np.isfinite(kl).all())
    self.assertTrue(np.isfinite(grads).all())


if __name__ == '__main__':
  absltest.main()



# Copyright 2021 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Utility math functions."""

import functools
from typing import Optional, Tuple

import chex
import jax
from jax import core as jax_core
from jax.custom_derivatives import SymbolicZero
import jax.numpy as jnp

Array = chex.Array


@jax.custom_jvp
def multiply_no_nan(x: Array, y: Array) -> Array:
  """Equivalent of TF `multiply_no_nan`.

  Computes the element-wise product of `x` and `y` and return 0 if `y` is zero,
  even if `x` is NaN or infinite.

  Args:
    x: First input.
    y: Second input.

  Returns:
    The product of `x` and `y`.

  Raises:
    ValueError if the shapes of `x` and `y` do not match.
  """
  dtype = jnp.result_type(x, y)
  return jnp.where(y == 0, jnp.zeros((), dtype=dtype), x * y)


# TODO(dougalm): move helpers like these into JAX AD utils
def add_maybe_symbolic(x, y):
  if isinstance(x, SymbolicZero):
    return y
  elif isinstance(y, SymbolicZero):
    return x
  else:
    return x + y


def scale_maybe_symbolic(result_aval, tangent, scale):
  if isinstance(tangent, SymbolicZero):
    return SymbolicZero(result_aval)
  else:
    return tangent * scale


@functools.partial(multiply_no_nan.defjvp, symbolic_zeros=True)
def multiply_no_nan_jvp(
    primals: Tuple[Array, Array],
    tangents: Tuple[Array, Array]) -> Tuple[Array, Array]:
  """Custom gradient computation for `multiply_no_nan`."""
  x, y = primals
  x_dot, y_dot = tangents
  primal_out = multiply_no_nan(x, y)
  primal_aval = jax_core.get_aval(primal_out)
  result_aval = primal_aval.at_least_vspace()
  tangent_out_1 = scale_maybe_symbolic(result_aval, x_dot, y)
  tangent_out_2 = scale_maybe_symbolic(result_aval, y_dot, x)
  return primal_out, add_maybe_symbolic(tangent_out_1, tangent_out_2)


@jax.custom_jvp
def power_no_nan(x: Array, y: Array) -> Array:
  """Computes `x ** y` and ensure that the result is 1.0 when `y` is zero.

  Compute the element-wise power `x ** y` and return 1.0 when `y` is zero,
  regardless of the value of `x`, even if it is NaN or infinite. This method
  uses the convention `0 ** 0 = 1`.

  Args:
    x: First input.
    y: Second input.

  Returns:
    The power `x ** y`.
  """
  dtype = jnp.result_type(x, y)
  return jnp.where(y == 0, jnp.ones((), dtype=dtype), jnp.power(x, y))


@power_no_nan.defjvp
def power_no_nan_jvp(
    primals: Tuple[Array, Array],
    tangents: Tuple[Array, Array]) -> Tuple[Array, Array]:
  """Custom gradient computation for `power_no_nan`."""
  x, y = primals
  x_dot, y_dot = tangents
  primal_out = power_no_nan(x, y)
  tangent_out = (y * power_no_nan(x, y - 1) * x_dot
                 + primal_out * jnp.log(x) * y_dot)
  return primal_out, tangent_out


def mul_exp(x: Array, logp: Array) -> Array:
  """Returns `x * exp(logp)` with zero output if `exp(logp)==0`.

  Args:
    x: An array.
    logp: An array.

  Returns:
    `x * exp(logp)` with zero output and zero gradient if `exp(logp)==0`,
    even if `x` is NaN or infinite.
  """
  p = jnp.exp(logp)
  # If p==0, the gradient with respect to logp is zero,
  # so we can replace the possibly non-finite `x` with zero.
  x = jnp.where(p == 0, 0.0, x)
  return x * p


def normalize(
    *, probs: Optional[Array] = None, logits: Optional[Array] = None) -> Array:
  """Normalize logits (via log_softmax) or probs (ensuring they sum to one)."""
  if logits is None:
    probs = jnp.asarray(probs)
    return probs / probs.sum(axis=-1, keepdims=True)
  else:
    logits = jnp.asarray(logits)
    return jax.nn.log_softmax(logits, axis=-1)


def sum_last(x: Array, ndims: int) -> Array:
  """Sums the last `ndims` axes of array `x`."""
  axes_to_sum = tuple(range(-ndims, 0))
  return jnp.sum(x, axis=axes_to_sum)


def log_expbig_minus_expsmall(big: Array, small: Array) -> Array:
  """Stable implementation of `log(exp(big) - exp(small))`.

  Args:
    big: First input.
    small: Second input. It must be `small <= big`.

  Returns:
    The resulting `log(exp(big) - exp(small))`.
  """
  return big + jnp.log1p(-jnp.exp(small - big))


def log_beta(a: Array, b: Array) -> Array:
  """Obtains the log of the beta function `log B(a, b)`.

  Args:
    a: First input. It must be positive.
    b: Second input. It must be positive.

  Returns:
    The value `log B(a, b) = log Gamma(a) + log Gamma(b) - log Gamma(a + b)`,
    where `Gamma` is the Gamma function, obtained through stable computation of
    `log Gamma`.
  """
  return jax.lax.lgamma(a) + jax.lax.lgamma(b) - jax.lax.lgamma(a + b)


def log_beta_multivariate(a: Array) -> Array:
  """Obtains the log of the multivariate beta function `log B(a)`.

  Args:
    a: An array of length `K` containing positive values.

  Returns:
    The value
    `log B(a) = sum_{k=1}^{K} log Gamma(a_k) - log Gamma(sum_{k=1}^{K} a_k)`,
    where `Gamma` is the Gamma function, obtained through stable computation of
    `log Gamma`.
  """
  return (
      jnp.sum(jax.lax.lgamma(a), axis=-1) - jax.lax.lgamma(jnp.sum(a, axis=-1)))



# Copyright 2021 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Tests for `uniform.py`."""

from absl.testing import absltest
from absl.testing import parameterized

import chex
from distrax._src.distributions import uniform
from distrax._src.utils import equivalence
import jax
import jax.numpy as jnp
import numpy as np


class UniformTest(equivalence.EquivalenceTest):

  def setUp(self):
    super().setUp()
    self._init_distr_cls(uniform.Uniform)

  @parameterized.named_parameters(
      ('1d', (0., 1.)),
      ('2d', (np.zeros(2), np.ones(2))),
      ('rank 2', (np.zeros((3, 2)), np.ones((3, 2)))),
      ('broadcasted low', (0., np.ones(3))),
      ('broadcasted high', (np.ones(3), 1.)),
  )
  def test_event_shape(self, distr_params):
    super()._test_event_shape(distr_params, dict())

  @chex.all_variants
  @parameterized.named_parameters(
      ('1d, no shape', (0., 1.), ()),
      ('1d, int shape', (0., 1.), 1),
      ('1d, 1-tuple shape', (0., 1.), (1,)),
      ('1d, 2-tuple shape', (0., 1.), (2, 2)),
      ('2d, no shape', (np.zeros(2), np.ones(2)), ()),
      ('2d, int shape', (np.zeros(2), np.ones(2)), 1),
      ('2d, 1-tuple shape', (np.zeros(2), np.ones(2)), (1,)),
      ('2d, 2-tuple shape', (np.zeros(2), np.ones(2)), (2, 2)),
      ('rank 2, 2-tuple shape', (np.zeros((3, 2)), np.ones((3, 2))), (2, 2)),
      ('broadcasted low', (0., np.ones(3)), (2, 2)),
      ('broadcasted high', (np.ones(3), 1.), ()),
  )
  def test_sample_shape(self, distr_params, sample_shape):
    super()._test_sample_shape(distr_params, dict(), sample_shape)

  @chex.all_variants
  @jax.numpy_rank_promotion('raise')
  @parameterized.named_parameters(
      ('1d, no shape', (0., 1.), ()),
      ('1d, int shape', (0., 1.), 1),
      ('1d, 1-tuple shape', (0., 1.), (1,)),
      ('1d, 2-tuple shape', (0., 1.), (2, 2)),
      ('2d, no shape', (np.zeros(2), np.ones(2)), ()),
      ('2d, int shape', (np.zeros(2), np.ones(2)), 1),
      ('2d, 1-tuple shape', (np.zeros(2), np.ones(2)), (1,)),
      ('2d, 2-tuple shape', (np.zeros(2), np.ones(2)), (2, 2)),
      ('rank 2, 2-tuple shape', (np.zeros((3, 2)), np.ones((3, 2))), (2, 2)),
      ('broadcasted low', (0., np.ones(3)), (2, 2)),
      ('broadcasted high', (np.ones(3), 1.), ()),
  )
  def test_sample_and_log_prob(self, distr_params, sample_shape):
    super()._test_sample_and_log_prob(
        dist_args=distr_params,
        dist_kwargs=dict(),
        sample_shape=sample_shape,
        assertion_fn=self.assertion_fn(rtol=1e-3))

  @chex.all_variants
  @parameterized.named_parameters(
      ('log_prob', 'log_prob'),
      ('prob', 'prob'),
      ('cdf', 'cdf'),
      ('survival_function', 'survival_function'),
      ('log_survival_function', 'log_survival_function')
  )
  def test_method_with_inputs(self, function_string):
    inputs = 10. * np.random.normal(size=(100,))
    super()._test_attribute(
        function_string, dist_args=(-1, 1), call_args=(inputs,))

  @chex.all_variants(with_pmap=False)
  @parameterized.named_parameters(
      ('entropy', (0., 1.), 'entropy'),
      ('mean', (0, 1), 'mean'),
      ('variance', (0, 1), 'variance'),
      ('variance from 1d params', (np.ones(2), np.ones(2)), 'mean'),
      ('stddev', (0, 1), 'stddev'),
      ('stddev from rank 2 params', (np.ones((2, 3)), np.ones(
          (2, 3))), 'stddev'),
  )
  def test_method(self, distr_params, function_string):
    super()._test_attribute(function_string, distr_params)

  @parameterized.named_parameters(
      ('low', 'low'),
      ('high', 'high'),
  )
  def test_attribute(self, attribute_string):
    super()._test_attribute(attribute_string)

  @chex.all_variants(with_pmap=False)
  def test_median(self):
    np.testing.assert_allclose(
        self.variant(self.distrax_cls(-1, 1).median)(), 0)

  @chex.all_variants(with_pmap=False)
  @parameterized.named_parameters(
      ('kl distrax_to_distrax', 'kl_divergence', 'distrax_to_distrax'),
      ('kl distrax_to_tfp', 'kl_divergence', 'distrax_to_tfp'),
      ('kl tfp_to_distrax', 'kl_divergence', 'tfp_to_distrax'),
      ('cross-ent distrax_to_distrax', 'cross_entropy', 'distrax_to_distrax'),
      ('cross-ent distrax_to_tfp', 'cross_entropy', 'distrax_to_tfp'),
      ('cross-ent tfp_to_distrax', 'cross_entropy', 'tfp_to_distrax'))
  def test_with_two_distributions(self, function_string, mode_string):
    super()._test_with_two_distributions(
        attribute_string=function_string,
        mode_string=mode_string,
        dist1_kwargs={
            'low': -0.5 + np.random.rand(4, 1, 2),
            'high': np.array([[1.8, 1.5], [1.1, 1.2], [1.4, 1.1]]),
        },
        dist2_kwargs={
            'low': -1.0 + np.random.rand(3, 2),
            'high': 1.5 + np.random.rand(4, 1, 2),
        },
        assertion_fn=self.assertion_fn(rtol=1e-3))

  def test_jittable(self):
    super()._test_jittable((0.0, 1.0))

  @parameterized.named_parameters(
      ('single element', 2),
      ('range', slice(-1)),
      ('range_2', (slice(None), slice(-1))),
      ('ellipsis', (Ellipsis, -1)),
  )
  def test_slice(self, slice_):
    low = jnp.zeros((3, 4, 5))
    high = jnp.ones((3, 4, 5))
    dist = self.distrax_cls(low=low, high=high)
    self.assertion_fn(rtol=1e-3)(dist[slice_].low, low[slice_])
    self.assertion_fn(rtol=1e-3)(dist[slice_].high, high[slice_])

  def test_slice_different_parameterization(self):
    low = jnp.zeros((3, 4, 5))
    high = 1.
    dist = self.distrax_cls(low=low, high=high)
    self.assertion_fn(rtol=1e-3)(dist[..., -1].low, low[..., -1])
    self.assertEqual(dist[..., -1].high.shape, (3, 4))
    self.assertion_fn(rtol=1e-3)(dist[..., -1].high, high)  # Not slicing high.

if __name__ == '__main__':
  absltest.main()



# Copyright 2021 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Tests for `mvn_diag_plus_low_rank.py`."""

from absl.testing import absltest
from absl.testing import parameterized

import chex
from distrax._src.distributions.mvn_diag_plus_low_rank import MultivariateNormalDiagPlusLowRank
from distrax._src.utils import equivalence

import jax
import jax.experimental
import jax.numpy as jnp
import numpy as np
from tensorflow_probability.substrates import jax as tfp

tfd = tfp.distributions


def _covariance_matrix_from_low_rank(
    scale_diag, scale_u_matrix, scale_v_matrix):
  """Constructs the covariance matrix from the low-rank matrices."""
  if scale_u_matrix is not None:
    if scale_v_matrix is None:
      scale_v_matrix = np.copy(scale_u_matrix)
    scale_v_matrix_t = np.vectorize(
        np.transpose, signature='(k,m)->(m,k)')(scale_v_matrix)
    scale = np.matmul(scale_u_matrix, scale_v_matrix_t) + np.vectorize(
        np.diag, signature='(k)->(k,k)')(scale_diag)
  else:
    scale = np.vectorize(np.diag, signature='(k)->(k,k)')(scale_diag)
  scale_t = np.vectorize(np.transpose, signature='(k,k)->(k,k)')(scale)
  return np.matmul(scale, scale_t)


class MultivariateNormalDiagPlusLowRankTest(equivalence.EquivalenceTest):

  def setUp(self):
    super().setUp()
    self._init_distr_cls(MultivariateNormalDiagPlusLowRank)

  @parameterized.named_parameters(
      ('all inputs are None', {}),
      ('scale_v_matrix is provided but scale_u_matrix is None', {
          'loc': np.zeros((4,)),
          'scale_v_matrix': np.ones((4, 2)),
      }),
      ('wrong dimension of loc', {
          'loc': np.array(0.),
      }),
      ('wrong dimension of scale_diag', {
          'scale_diag': np.array(0.),
      }),
      ('wrong dimension of scale_u_matrix', {
          'scale_u_matrix': np.ones((4,)),
      }),
      ('wrong dimension of scale_v_matrix', {
          'scale_u_matrix': np.ones((4, 2)),
          'scale_v_matrix': np.ones((4,)),
      }),
      ('last dimension of scale_u_matrix is zero', {
          'scale_u_matrix': np.ones((4, 0)),
      }),
      ('inconsistent dimensions of scale_u_matrix and scale_v_matrix', {
          'scale_u_matrix': np.ones((4, 2)),
          'scale_v_matrix': np.ones((4, 1)),
      }),
      ('inconsistent event_dim across two params', {
          'loc': np.zeros((4,)),
          'scale_u_matrix': np.ones((5, 2)),
      }),
      ('inconsistent event_dim across three params', {
          'loc': np.zeros((4,)),
          'scale_diag': np.ones((5,)),
          'scale_u_matrix': np.ones((4, 2)),
      }),
  )
  def test_raises_on_wrong_inputs(self, dist_kwargs):
    with self.assertRaises(ValueError):
      MultivariateNormalDiagPlusLowRank(**dist_kwargs)

  @parameterized.named_parameters(
      ('loc provided', {'loc': np.zeros((4,))}),
      ('scale_diag provided', {'scale_diag': np.ones((4,))}),
      ('scale_u_matrix provided', {'scale_u_matrix': np.zeros((4, 1))}),
  )
  def test_default_properties(self, dist_kwargs):
    dist = MultivariateNormalDiagPlusLowRank(**dist_kwargs)
    self.assertion_fn(rtol=1e-3)(dist.loc, jnp.zeros((4,)))
    self.assertion_fn(rtol=1e-3)(dist.scale_diag, jnp.ones((4,)))
    self.assertion_fn(rtol=1e-3)(dist.scale_u_matrix, jnp.zeros((4, 1)))
    self.assertion_fn(rtol=1e-3)(dist.scale_v_matrix, jnp.zeros((4, 1)))

  @parameterized.named_parameters(
      ('unbatched', (), (4,), (4,), (4, 2), (4, 2)),
      ('batched loc', (7,), (7, 4), (4,), (4, 2), (4, 2)),
      ('batched scale_diag', (7,), (4,), (7, 4), (4, 2), (4, 2)),
      ('batched scale_u_matrix', (7,), (4,), (4,), (7, 4, 2), (4, 2)),
      ('batched scale_v_matrix', (7,), (4,), (4,), (4, 2), (7, 4, 2)),
  )
  def test_properties(self, batch_shape, loc_shape, scale_diag_shape,
                      scale_u_matrix_shape, scale_v_matrix_shape):
    rng = np.random.default_rng(2022)
    loc = rng.normal(size=loc_shape)
    scale_diag = rng.normal(size=scale_diag_shape)
    scale_u_matrix = rng.normal(size=scale_u_matrix_shape)
    scale_v_matrix = rng.normal(size=scale_v_matrix_shape)
    dist = MultivariateNormalDiagPlusLowRank(
        loc=loc,
        scale_diag=scale_diag,
        scale_u_matrix=scale_u_matrix,
        scale_v_matrix=scale_v_matrix,
    )
    self.assertEqual(dist.batch_shape, batch_shape)
    self.assertion_fn(rtol=1e-3)(
        dist.loc, jnp.broadcast_to(loc, batch_shape + (4,)))
    self.assertion_fn(rtol=1e-3)(
        dist.scale_diag, jnp.broadcast_to(scale_diag, batch_shape + (4,)))
    self.assertion_fn(rtol=1e-3)(
        dist.scale_u_matrix,
        jnp.broadcast_to(scale_u_matrix, batch_shape + (4, 2)))
    self.assertion_fn(rtol=1e-3)(
        dist.scale_v_matrix,
        jnp.broadcast_to(scale_v_matrix, batch_shape + (4, 2)))

  @chex.all_variants
  @parameterized.named_parameters(
      ('unbatched, no shape', (), (4,), (4,), (4, 2), (4, 2)),
      ('batched loc, no shape', (), (7, 4), (4,), (4, 2), (4, 2)),
      ('batched scale_diag, no shape', (), (4,), (7, 4), (4, 2), (4, 2)),
      ('batched scale_u_matrix, no shape', (), (4,), (4,), (7, 4, 2), (4, 2)),
      ('batched scale_v_matrix, no shape', (), (4,), (4,), (4, 2), (7, 4, 2)),
      ('unbatched, with shape', (3,), (4,), (4,), (4, 2), (4, 2)),
      ('batched loc, with shape', (3,), (7, 4), (4,), (4, 2), (4, 2)),
      ('batched scale_diag, with shape', (3,), (4,), (7, 4), (4, 2), (4, 2)),
      ('batched scale_u_matrix, with shape',
       (3,), (4,), (4,), (7, 4, 2), (4, 2)),
      ('batched scale_v_matrix, with shape',
       (3,), (4,), (4,), (4, 2), (7, 4, 2)),
  )
  def test_sample_shape(self, sample_shape, loc_shape, scale_diag_shape,
                        scale_u_matrix_shape, scale_v_matrix_shape):
    rng = np.random.default_rng(2022)
    loc = rng.normal(size=loc_shape)
    scale_diag = rng.normal(size=scale_diag_shape)
    scale_u_matrix = rng.normal(size=scale_u_matrix_shape)
    scale_v_matrix = rng.normal(size=scale_v_matrix_shape)
    dist = MultivariateNormalDiagPlusLowRank(
        loc=loc,
        scale_diag=scale_diag,
        scale_u_matrix=scale_u_matrix,
        scale_v_matrix=scale_v_matrix,
    )
    tfp_dist = tfd.MultivariateNormalFullCovariance(
        loc=loc,
        covariance_matrix=_covariance_matrix_from_low_rank(
            scale_diag, scale_u_matrix, scale_v_matrix)
    )
    sample_fn = self.variant(
        lambda rng: dist.sample(sample_shape=sample_shape, seed=rng))
    distrax_samples = sample_fn(jax.random.PRNGKey(0))
    tfp_samples = tfp_dist.sample(
        sample_shape=sample_shape, seed=jax.random.PRNGKey(0))
    self.assertEqual(distrax_samples.shape, tfp_samples.shape)

  @chex.all_variants
  @parameterized.named_parameters(
      ('float32', jnp.float32),
      ('float64', jnp.float64))
  def test_sample_dtype(self, dtype):
    with jax.experimental.enable_x64(dtype.dtype.itemsize == 8):
      dist_params = {
          'loc': np.array([0., 0.], dtype),
          'scale_diag': np.array([1., 1.], dtype)}
      dist = MultivariateNormalDiagPlusLowRank(**dist_params)
      samples = self.variant(dist.sample)(seed=jax.random.PRNGKey(0))
      self.assertEqual(samples.dtype, dist.dtype)
      chex.assert_type(samples, dtype)

  @chex.all_variants
  @parameterized.named_parameters(
      ('unbatched, unbatched value', (4,), (4,), (4,), (4, 2), (4, 2)),
      ('batched loc, unbatched value', (4,), (7, 4), (4,), (4, 2), (4, 2)),
      ('batched scale_diag, unbatched value',
       (4,), (4,), (7, 4), (4, 2), (4, 2)),
      ('batched scale_u_matrix, unbatched value',
       (4,), (4,), (4,), (7, 4, 2), (4, 2)),
      ('batched scale_v_matrix, unbatched value',
       (4,), (4,), (4,), (4, 2), (7, 4, 2)),
      ('unbatched, batched value', (7, 4), (4,), (4,), (4, 2), (4, 2)),
      ('batched loc, batched value', (7, 4), (7, 4), (4,), (4, 2), (4, 2)),
      ('batched scale_diag, batched value',
       (7, 4), (4,), (7, 4), (4, 2), (4, 2)),
      ('batched scale_u_matrix, batched value',
       (7, 4), (4,), (4,), (7, 4, 2), (4, 2)),
      ('batched scale_v_matrix, batched value',
       (7, 4), (4,), (4,), (4, 2), (7, 4, 2)),
  )
  def test_log_prob(self, value_shape, loc_shape, scale_diag_shape,
                    scale_u_matrix_shape, scale_v_matrix_shape):
    rng = np.random.default_rng(2022)
    loc = rng.normal(size=loc_shape)
    scale_diag = rng.normal(size=scale_diag_shape)
    scale_u_matrix = 0.1 * rng.normal(size=scale_u_matrix_shape)
    scale_v_matrix = 0.1 * rng.normal(size=scale_v_matrix_shape)
    dist = MultivariateNormalDiagPlusLowRank(
        loc=loc,
        scale_diag=scale_diag,
        scale_u_matrix=scale_u_matrix,
        scale_v_matrix=scale_v_matrix,
    )
    tfp_dist = tfd.MultivariateNormalFullCovariance(
        loc=loc,
        covariance_matrix=_covariance_matrix_from_low_rank(
            scale_diag, scale_u_matrix, scale_v_matrix)
    )
    value = rng.normal(size=value_shape)
    self.assertion_fn(rtol=2e-3)(
        self.variant(dist.log_prob)(value), tfp_dist.log_prob(value))

  @chex.all_variants(with_pmap=False)
  @parameterized.named_parameters(
      ('unbatched', (4,), (4,), (4, 2), (4, 2)),
      ('batched loc', (7, 4), (4,), (4, 2), (4, 2)),
      ('batched scale_diag', (4,), (7, 4), (4, 2), (4, 2)),
      ('batched scale_u_matrix', (4,), (4,), (7, 4, 2), (4, 2)),
      ('batched scale_v_matrix', (4,), (4,), (4, 2), (7, 4, 2)),
      ('scale_u_matrix is None', (4,), (4,), None, None),
      ('scale_v_matrix is None', (4,), (4,), (4, 2), None),
  )
  def test_method(self, loc_shape, scale_diag_shape,
                  scale_u_matrix_shape, scale_v_matrix_shape):
    rng = np.random.default_rng(2022)
    loc = rng.normal(size=loc_shape)
    scale_diag = rng.normal(size=scale_diag_shape)
    if scale_u_matrix_shape is None:
      scale_u_matrix = None
    else:
      scale_u_matrix = 0.1 * rng.normal(size=scale_u_matrix_shape)
    if scale_v_matrix_shape is None:
      scale_v_matrix = None
    else:
      scale_v_matrix = 0.1 * rng.normal(size=scale_v_matrix_shape)
    dist = MultivariateNormalDiagPlusLowRank(
        loc=loc,
        scale_diag=scale_diag,
        scale_u_matrix=scale_u_matrix,
        scale_v_matrix=scale_v_matrix,
    )
    tfp_dist = tfd.MultivariateNormalFullCovariance(
        loc=loc,
        covariance_matrix=_covariance_matrix_from_low_rank(
            scale_diag, scale_u_matrix, scale_v_matrix)
    )
    for method in [
        'entropy', 'mean', 'stddev', 'variance', 'covariance', 'mode']:
      if method in ['stddev', 'variance']:
        rtol = 1e-2
      elif method in ['covariance']:
        rtol = 8e-2
      else:
        rtol = 1e-3
      with self.subTest(method=method):
        fn = self.variant(getattr(dist, method))
        self.assertion_fn(rtol=rtol)(fn(), getattr(tfp_dist, method)())

  @chex.all_variants(with_pmap=False)
  @parameterized.named_parameters(
      ('kl distrax_to_distrax', 'kl_divergence', 'distrax_to_distrax'),
      ('kl distrax_to_tfp', 'kl_divergence', 'distrax_to_tfp'),
      ('kl tfp_to_distrax', 'kl_divergence', 'tfp_to_distrax'),
      ('cross-ent distrax_to_distrax', 'cross_entropy', 'distrax_to_distrax'),
      ('cross-ent distrax_to_tfp', 'cross_entropy', 'distrax_to_tfp'),
      ('cross-ent tfp_to_distrax', 'cross_entropy', 'tfp_to_distrax'))
  def test_with_two_distributions(self, function_string, mode_string):
    rng = np.random.default_rng(2022)
    loc1 = rng.normal(size=(5, 1, 4))
    scale_diag1 = rng.normal(size=(3, 4))
    scale_u_matrix1 = 0.1 * rng.normal(size=(5, 1, 4, 2))
    scale_perturb_diag1 = rng.normal(size=(5, 1, 2))
    scale_v_matrix1 = scale_u_matrix1 * np.expand_dims(
        scale_perturb_diag1, axis=-2)
    loc2 = rng.normal(size=(3, 4))
    scale_diag2 = rng.normal(size=(3, 4))
    scale_u_matrix2 = 0.1 * rng.normal(size=(4, 2))
    scale_perturb_diag2 = rng.normal(size=(2,))
    scale_v_matrix2 = scale_u_matrix2 * np.expand_dims(
        scale_perturb_diag2, axis=-2)
    distrax_dist1 = MultivariateNormalDiagPlusLowRank(
        loc=loc1,
        scale_diag=scale_diag1,
        scale_u_matrix=scale_u_matrix1,
        scale_v_matrix=scale_v_matrix1,
    )
    distrax_dist2 = MultivariateNormalDiagPlusLowRank(
        loc=loc2,
        scale_diag=scale_diag2,
        scale_u_matrix=scale_u_matrix2,
        scale_v_matrix=scale_v_matrix2,
    )
    tfp_dist1 = tfd.MultivariateNormalDiagPlusLowRank(
        loc=loc1,
        scale_diag=scale_diag1,
        scale_perturb_factor=scale_u_matrix1,
        scale_perturb_diag=scale_perturb_diag1,
    )
    tfp_dist2 = tfd.MultivariateNormalDiagPlusLowRank(
        loc=loc2,
        scale_diag=scale_diag2,
        scale_perturb_factor=scale_u_matrix2,
        scale_perturb_diag=scale_perturb_diag2,
    )
    expected_result1 = getattr(tfp_dist1, function_string)(tfp_dist2)
    expected_result2 = getattr(tfp_dist2, function_string)(tfp_dist1)
    if mode_string == 'distrax_to_distrax':
      result1 = self.variant(getattr(distrax_dist1, function_string))(
          distrax_dist2)
      result2 = self.variant(getattr(distrax_dist2, function_string))(
          distrax_dist1)
    elif mode_string == 'distrax_to_tfp':
      result1 = self.variant(getattr(distrax_dist1, function_string))(tfp_dist2)
      result2 = self.variant(getattr(distrax_dist2, function_string))(tfp_dist1)
    elif mode_string == 'tfp_to_distrax':
      result1 = self.variant(getattr(tfp_dist1, function_string))(distrax_dist2)
      result2 = self.variant(getattr(tfp_dist2, function_string))(distrax_dist1)
    else:
      raise ValueError(f'Unsupported mode: {mode_string}')
    self.assertion_fn(rtol=3e-3)(result1, expected_result1)
    self.assertion_fn(rtol=3e-3)(result2, expected_result2)

  def test_jittable(self):
    super()._test_jittable(
        dist_kwargs={'loc': np.zeros((4,))},
        assertion_fn=self.assertion_fn(rtol=1e-3))

  @parameterized.named_parameters(
      ('single element', 2),
      ('range', slice(-1)),
      ('range_2', (slice(None), slice(-1))),
  )
  def test_slice(self, slice_):
    rng = np.random.default_rng(2022)
    loc = rng.normal(size=(6, 5, 4))
    scale_diag = rng.normal(size=(4,))
    scale_u_matrix = rng.normal(size=(1, 4, 2))
    scale_v_matrix = rng.normal(size=(4, 2))
    dist = MultivariateNormalDiagPlusLowRank(
        loc=loc,
        scale_diag=scale_diag,
        scale_u_matrix=scale_u_matrix,
        scale_v_matrix=scale_v_matrix,
    )
    self.assertEqual(dist[slice_].batch_shape, loc[slice_].shape[:-1])
    self.assertEqual(dist[slice_].event_shape, loc[slice_].shape[-1:])
    self.assertion_fn(rtol=1e-3)(dist[slice_].mean(), loc[slice_])

  def test_slice_ellipsis(self):
    rng = np.random.default_rng(2022)
    loc = rng.normal(size=(6, 5, 4))
    scale_diag = rng.normal(size=(4,))
    scale_u_matrix = rng.normal(size=(1, 4, 2))
    scale_v_matrix = rng.normal(size=(4, 2))
    dist = MultivariateNormalDiagPlusLowRank(
        loc=loc,
        scale_diag=scale_diag,
        scale_u_matrix=scale_u_matrix,
        scale_v_matrix=scale_v_matrix,
    )
    self.assertEqual(dist[..., -1].batch_shape, (6,))
    self.assertEqual(dist[..., -1].event_shape, dist.event_shape)
    self.assertion_fn(rtol=1e-3)(dist[..., -1].mean(), loc[:, -1, :])


if __name__ == '__main__':
  absltest.main()



# Copyright 2021 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Epsilon-Greedy distributions with respect to a set of preferences."""

from typing import Any, Union

import chex
from distrax._src.distributions import categorical
from distrax._src.distributions import distribution
import jax.numpy as jnp


Array = chex.Array


def _argmax_with_random_tie_breaking(preferences: Array) -> Array:
  """Compute probabilities greedily with respect to a set of preferences."""
  optimal_actions = preferences == preferences.max(axis=-1, keepdims=True)
  return optimal_actions / optimal_actions.sum(axis=-1, keepdims=True)


def _mix_probs_with_uniform(probs: Array, epsilon: float) -> Array:
  """Mix an arbitrary categorical distribution with a uniform distribution."""
  num_actions = probs.shape[-1]
  uniform_probs = jnp.ones_like(probs) / num_actions
  return (1 - epsilon) * probs + epsilon * uniform_probs


class EpsilonGreedy(categorical.Categorical):
  """A Categorical that is ε-greedy with respect to some preferences.

  Given a set of unnormalized preferences, the distribution is a mixture
  of the Greedy and Uniform distribution; with weight (1-ε) and ε, respectively.
  """

  def __init__(self,
               preferences: Array,
               epsilon: float,
               dtype: Union[jnp.dtype, type[Any]] = int):
    """Initializes an EpsilonGreedy distribution.

    Args:
      preferences: Unnormalized preferences.
      epsilon: Mixing parameter ε.
      dtype: The type of event samples.
    """
    self._preferences = jnp.asarray(preferences)
    self._epsilon = epsilon
    greedy_probs = _argmax_with_random_tie_breaking(self._preferences)
    probs = _mix_probs_with_uniform(greedy_probs, epsilon)
    super().__init__(probs=probs, dtype=dtype)

  @property
  def epsilon(self) -> float:
    """Mixing parameters of the distribution."""
    return self._epsilon

  @property
  def preferences(self) -> Array:
    """Unnormalized preferences."""
    return self._preferences

  def __getitem__(self, index) -> 'EpsilonGreedy':
    """See `Distribution.__getitem__`."""
    index = distribution.to_batch_shape_index(self.batch_shape, index)
    return EpsilonGreedy(
        preferences=self.preferences[index],
        epsilon=self.epsilon,
        dtype=self.dtype)
